from ..objects._base import WorldObject
from ..cameras import PerspectiveCamera
from ..resources import Texture
from ..renderers import WgpuRenderer
from ..renderers.wgpu import GfxTextureView


def _is_cube_texture(texture):
    if not isinstance(texture, Texture):
        return False
    return texture.dim == 2 and len(texture.size) == 3 and texture.size[2] == 6


class _CubeCameraRenderer(WgpuRenderer):
    def __init__(self, target):
        assert _is_cube_texture(target), "target must be a cube texture"

        super().__init__(target)

        # Pre generate views of different layers of the cube texture
        self._target_views = []
        for layer in range(6):
            self._target_views.append(
                GfxTextureView(
                    target,
                    view_dim="2d",
                    layer_range=(layer, layer + 1),
                    mip_range=(0, 1),
                )
            )

    def flush(self, layer):
        super().flush(self._target_views[layer])


class CubeCamera(WorldObject):
    """
    Create a camera array to help create a cube texture map from a viewpoint in a scene.

    Note that the texture data will be written directly to the internal "GPUTexture" object,
    not to the "data" attribute of target. That is, its data cannot be accessed from the CPU.
    """

    def __init__(self, target, near=0.1, far=1000):
        super().__init__()

        self._renderer = _CubeCameraRenderer(target)

        fov = 90
        aspect = 1
        depth_range = near, far

        # By convention, cube maps are specified in a coordinate system in which positive-x is to the right when looking at the positive-z axis,
        # that is, it using a left-handed coordinate system.
        # Since gfx uses a right-handed coordinate system, environment maps used in gfx will have pos-x and neg-x swapped.
        # so camrea_px is actually looking at the neg-x direction, and camera_nx is looking at the pos-x direction.

        camera_px = PerspectiveCamera(fov, aspect, depth_range=depth_range)
        camera_px.world.reference_up = (0, 1, 0)
        camera_px.look_at((-1, 0, 0))
        self.add(camera_px)

        camera_nx = PerspectiveCamera(fov, aspect, depth_range=depth_range)
        camera_nx.world.reference_up = (0, 1, 0)
        camera_nx.look_at((1, 0, 0))
        self.add(camera_nx)

        camera_py = PerspectiveCamera(fov, aspect, depth_range=depth_range)
        camera_py.world.reference_up = (0, 0, -1)
        camera_py.look_at((0, 1, 0))
        self.add(camera_py)

        camera_ny = PerspectiveCamera(fov, aspect, depth_range=depth_range)
        camera_ny.world.reference_up = (0, 0, 1)
        camera_ny.look_at((0, -1, 0))
        self.add(camera_ny)

        camera_pz = PerspectiveCamera(fov, aspect, depth_range=depth_range)
        camera_pz.world.reference_up = (0, 1, 0)
        camera_pz.look_at((0, 0, 1))
        self.add(camera_pz)

        camera_nz = PerspectiveCamera(fov, aspect, depth_range=depth_range)
        camera_nz.world.reference_up = (0, 1, 0)
        camera_nz.look_at((0, 0, -1))
        self.add(camera_nz)

    @property
    def renderer(self):
        """The renderer used to render the scene to the cube texture."""

        return self._renderer

    def render(self, scene):
        """Render the scene from the cube camera's perspective, and write the result to the target texture."""

        renderer = self.renderer
        cameras = self.children
        assert len(cameras) == 6

        for layer, camera in enumerate(cameras):
            renderer.render(scene, camera, flush=False)
            renderer.flush(layer)



import math
import numpy as np
import wgpu  # only for flags/enums


from ....objects import Mesh, InstancedMesh, SkinnedMesh, WorldObject, Line, Points
from ....resources import Buffer, Texture, TextureMap
from ....utils import normals_from_vertices
from ....materials import (
    MeshBasicMaterial,
    MeshPhongMaterial,
    MeshToonMaterial,
    MeshNormalMaterial,
    MeshNormalLinesMaterial,
    MeshSliceMaterial,
    MeshStandardMaterial,
    MeshPhysicalMaterial,
)

from .. import (
    register_wgpu_render_function,
    BaseShader,
    Binding,
    nchannels_from_format,
    to_texture_format,
    GfxSampler,
    GfxTextureView,
    load_wgsl,
)


@register_wgpu_render_function(WorldObject, MeshBasicMaterial)
class MeshShader(BaseShader):
    type = "render"

    def __init__(self, wobject):
        super().__init__(wobject)

        material = wobject.material
        geometry = wobject.geometry

        # Is this an instanced mesh?
        self["instanced"] = isinstance(wobject, InstancedMesh)

        # Is this a skinned mesh?
        self["use_skinning"] = isinstance(wobject, SkinnedMesh)

        # Is this a morphing mesh?
        morph_attrs = [
            getattr(geometry, name, None)
            for name in ["morph_positions", "morph_normals", "morph_colors"]
        ]
        morph_attrs = [x for x in morph_attrs if x is not None]
        self["use_morph_targets"] = bool(morph_attrs)

        # Is this a wireframe mesh?
        self["wireframe"] = getattr(material, "wireframe", False)
        self["flat_shading"] = getattr(material, "flat_shading", False)

        # Lighting off in the base class
        self["lighting"] = ""
        self["receive_shadow"] = wobject.receive_shadow

        # Per-vertex color, colormap, or a plane color?
        self["colorspace"] = "srgb"

        color_mode = str(material.color_mode).split(".")[-1]

        self["color_mode"] = color_mode
        self["use_uniform_color"] = color_mode in ("auto", "uniform")

        if material.map is not None and color_mode in (
            "vertex_map",
            "face_map",
            "auto",
        ):
            self["use_map"] = True
            self["colorspace"] = material.map.texture.colorspace
        else:
            self["use_map"] = False

        if getattr(geometry, "colors", None) and (
            color_mode in ("vertex", "face", "auto")
        ):
            nchannels = nchannels_from_format(geometry.colors.format)
            self["use_vertex_color"] = True
            self["color_buffer_channels"] = nchannels
            if nchannels not in (1, 2, 3, 4):
                raise ValueError(f"Geometry.colors needs 1-4 columns, not {nchannels}")
        else:
            self["color_buffer_channels"] = 0
            self["use_vertex_color"] = False

    def _define_texture_map(self, geometry, map, name, view_dim="2d", check=True):
        if check:
            # Check that the texture is compatible with the texcoord
            self._check_texture(map, geometry, view_dim)
        view = GfxTextureView(map.texture, view_dim=view_dim)

        filter_mode = f"{map.mag_filter}, {map.min_filter}, {map.mipmap_filter}"
        address_mode = f"{map.wrap_s}, {map.wrap_t}"
        sampler = GfxSampler(filter_mode, address_mode)

        self[f"{name}_uv"] = map.uv_channel

        bindings = [
            Binding(f"s_{name}", "sampler/filtering", sampler, "FRAGMENT"),
            Binding(f"t_{name}", "texture/auto", view, "FRAGMENT"),
        ]

        # todo: Only "UVMapping mode" have "transform" and "uv_channel"
        # todo: introduce "MappingMode" in TextureMap
        bindings.append(
            Binding(f"u_{name}", "buffer/uniform", map.uniform_buffer, "FRAGMENT"),
        )

        if map.uv_channel not in self["used_uv"]:
            texcoords = getattr(geometry, f"texcoords{map.uv_channel or ''}", None)
            if texcoords is not None:
                bindings.append(
                    Binding(
                        f"s_texcoords{map.uv_channel or ''}",
                        "buffer/read_only_storage",
                        texcoords,
                        "VERTEX",
                    )
                )

                if texcoords.data.ndim == 1:
                    self["used_uv"][map.uv_channel] = 1
                else:
                    self["used_uv"][map.uv_channel] = texcoords.data.shape[-1]

        return bindings

    def get_bindings(self, wobject, shared):
        geometry = wobject.geometry
        material = wobject.material

        # record all uv channels used by texture maps
        self["used_uv"] = {}  # {uv_channel: texcoords_dim}

        # We're assuming the presence of an index buffer for now
        assert getattr(geometry, "indices", None)

        # Triangles or quads?
        if geometry.indices.data is not None and geometry.indices.data.shape[-1] == 4:
            self["indexer"] = 6
        else:
            self["indexer"] = 3

        # Normals. Usually it'd be given. If not, we'll calculate it from the vertices.
        if getattr(geometry, "normals", None) is not None:
            normal_buffer = geometry.normals
        else:
            if geometry.indices.data.shape[-1] == 4:
                n1 = normals_from_vertices(
                    geometry.positions.data, geometry.indices.data[..., :3]
                )
                n2 = normals_from_vertices(
                    geometry.positions.data, geometry.indices.data[..., [0, 2, 3]]
                )
                normal_data = (n1 + n2) / 2
            else:
                normal_data = normals_from_vertices(
                    geometry.positions.data, geometry.indices.data
                )
            normal_buffer = Buffer(normal_data)

        # Init bindings
        rbuffer = "buffer/read_only_storage"
        bindings = [
            Binding("u_stdinfo", "buffer/uniform", shared.uniform_buffer),
            Binding("u_wobject", "buffer/uniform", wobject.uniform_buffer),
            Binding("u_material", "buffer/uniform", material.uniform_buffer),
            Binding("s_indices", rbuffer, geometry.indices, "VERTEX"),
            Binding("s_positions", rbuffer, geometry.positions, "VERTEX"),
            Binding("s_normals", rbuffer, normal_buffer, "VERTEX"),
        ]

        if self["use_vertex_color"]:
            bindings.append(Binding("s_colors", rbuffer, geometry.colors, "VERTEX"))

        if getattr(geometry, "tangents", None):
            bindings.append(Binding("s_tangents", rbuffer, geometry.tangents, "VERTEX"))
            self["use_tangent"] = True

        if self["use_map"]:
            map = material.map
            map_fmt, map_dim = to_texture_format(map.texture.format), map.texture.dim
            is_standard_map = (
                map_dim == 2
                and ("norm" in map_fmt or "float" in map_fmt)
                and ("rgb" in map_fmt)  # note: we assume r*, rg* maps are colormap
            )

            if not is_standard_map:
                # It's a 'generic' colormap
                self["use_colormap"] = True
                bindings.extend(
                    self.define_generic_colormap(material.map, geometry.texcoords)
                )
                if 0 not in self["used_uv"]:
                    texcoords = getattr(geometry, "texcoords", None)
                    bindings.append(
                        Binding("s_texcoords", rbuffer, texcoords, "VERTEX")
                    )
                    if texcoords.data.ndim == 1:
                        self["used_uv"][0] = 1
                    else:
                        self["used_uv"][0] = texcoords.data.shape[-1]
            else:
                # It's a classic mesh map
                bindings.extend(self._define_texture_map(geometry, material.map, "map"))

            self["colorspace"] = material.map.texture.colorspace

        if self["use_skinning"]:
            # Skinning requires skin_index and skin_weight buffers
            assert hasattr(geometry, "skin_indices") and hasattr(
                geometry, "skin_weights"
            )

            bindings.append(
                Binding("s_skin_indices", rbuffer, geometry.skin_indices, "VERTEX")
            )
            bindings.append(
                Binding("s_skin_weights", rbuffer, geometry.skin_weights, "VERTEX")
            )

            # Skinning requires a bone_matrices buffer

            skeleton = wobject.skeleton

            bindings.append(
                Binding(
                    "u_bone_matrices",
                    "buffer/uniform",
                    skeleton.bone_matrices_buffer,
                    "VERTEX",
                )
            )

        if self["use_morph_targets"]:
            # Get or create the morph texture and associated info
            try:
                morph_texture_info = geometry["_gfx_morph_texture"]
            except KeyError:
                morph_texture_info = self._encode_morph_texture(geometry, shared)
                geometry["_gfx_morph_texture"] = morph_texture_info
            morph_texture, stride, width, morph_count = morph_texture_info

            morph_target_influences = wobject._morph_target_influences  # buffer

            if morph_texture and morph_target_influences:
                view = GfxTextureView(morph_texture, view_dim="2d-array")
                bindings.append(
                    Binding("t_morph_targets", "texture/auto", view, "VERTEX")
                )

                self["influences_buffer_size"] = morph_target_influences.nitems
                self["morph_targets_count"] = morph_count
                self["morph_targets_stride"] = stride
                self["morph_targets_texture_width"] = width

            else:
                self["use_morph_targets"] = False

        # specular map configs, only for basic and phong materials now
        if getattr(material, "specular_map", None):
            bindings.extend(
                self._define_texture_map(
                    geometry, material.specular_map, "specular_map"
                )
            )
            self["use_specular_map"] = True

        # set emissive_map configs, for phong and standard materials
        if getattr(material, "emissive_map", None):
            bindings.extend(
                self._define_texture_map(
                    geometry, material.emissive_map, "emissive_map"
                )
            )
            self["use_emissive_map"] = True

        # set normal_map configs
        if getattr(material, "normal_map", None):
            bindings.extend(
                self._define_texture_map(geometry, material.normal_map, "normal_map")
            )
            self["use_normal_map"] = True

        # set envmap configs
        if getattr(material, "env_map", None):
            # TODO: Support envmap not only cube, but also equirect (hdr format)

            # special check for env_map
            assert isinstance(material.env_map, TextureMap)
            assert material.env_map.texture.size[2] == 6, "env_map must be a cube map"
            fmt = to_texture_format(material.env_map.texture.format)
            assert "norm" in fmt or "float" in fmt

            bindings.extend(
                self._define_texture_map(
                    geometry, material.env_map, "env_map", view_dim="cube", check=False
                )  # check=False because we don't need texcoords for env_map
            )

            if isinstance(material, MeshStandardMaterial):
                self["USE_IBL"] = True
            elif isinstance(material, MeshBasicMaterial):
                self["use_env_map"] = True
                self["env_combine_mode"] = getattr(
                    material, "env_combine_mode", "MULTIPLY"
                )

            self["env_mapping_mode"] = getattr(
                material, "env_mapping_mode", "CUBE-REFLECTION"
            )

        # set lightmap configs
        if getattr(material, "light_map", None):
            bindings.extend(
                self._define_texture_map(geometry, material.light_map, "light_map")
            )
            self["use_light_map"] = True

        # set aomap configs
        if getattr(material, "ao_map", None):
            bindings.extend(
                self._define_texture_map(geometry, material.ao_map, "ao_map")
            )
            self["use_ao_map"] = True

        # Define shader code for binding
        bindings = {i: binding for i, binding in enumerate(bindings)}
        self.define_bindings(0, bindings)

        # Instanced meshes have an extra storage buffer that we add manually
        bindings1 = {}  # non-auto-generated bindings
        if self["instanced"]:
            bindings1[0] = Binding(
                "s_instance_infos", rbuffer, wobject.instance_buffer, "VERTEX"
            )

        if self["use_morph_targets"]:
            bindings1[1] = Binding(
                "u_morph_target_influences",
                "buffer/uniform",
                morph_target_influences,
                "VERTEX",
            )

        return {
            0: bindings,
            1: bindings1,
        }

    def _encode_morph_texture(self, geometry, shared):
        morph_positions = getattr(geometry, "morph_positions", None)
        morph_normals = getattr(geometry, "morph_normals", None)
        morph_colors = getattr(geometry, "morph_colors", None)

        morph_attrs = [morph_positions, morph_normals, morph_colors]
        morph_attrs = [x for x in morph_attrs if x is not None]
        morph_count = min(len(x) for x in morph_attrs)

        vetex_data_count = 0

        if morph_positions:
            vetex_data_count = 1

        if morph_normals:
            vetex_data_count = 2

        if morph_colors:
            vetex_data_count = 3

        if vetex_data_count == 0:
            return None, None, None, None

        vertice_count = geometry.positions.nitems
        total_count = vertice_count * vetex_data_count
        width = total_count
        height = 1

        max_texture_width = shared.device.limits["max-texture-dimension-2d"]

        if width > max_texture_width:
            height = math.ceil(width / max_texture_width)
            width = max_texture_width

        buffer = np.zeros((morph_count, height * width, 4), dtype=np.float32)

        for i in range(morph_count):
            if morph_positions:
                morph_position = morph_positions[i]
                assert len(morph_position) == vertice_count, (
                    f"Morph target {i} has {len(morph_position)} vertices, expected {vertice_count}"
                )
                morph_position = np.pad(morph_position, ((0, 0), (0, 1)), "constant")
            else:
                morph_position = np.zeros((vertice_count, 4), dtype=np.float32)

            if morph_normals:
                morph_normal = morph_normals[i]
                assert len(morph_normal) == vertice_count, (
                    f"Morph normal {i} has {len(morph_normal)} vertices, expected {vertice_count}"
                )
                morph_normal = np.pad(morph_normal, ((0, 0), (0, 1)), "constant")
            else:
                morph_normal = np.zeros((vertice_count, 4), dtype=np.float32)

            if morph_colors:
                morph_color = morph_colors[i]
                assert len(morph_color) == vertice_count, (
                    f"Morph color {i} has {len(morph_colors[i])} vertices, expected {vertice_count}"
                )
            else:
                morph_color = np.zeros((vertice_count, 4), dtype=np.float32)

            morph_data = np.stack(
                (morph_position, morph_normal, morph_color)[:vetex_data_count], axis=1
            ).reshape(-1, 4)

            buffer[i, :total_count, :] = morph_data

        return (
            Texture(buffer, dim=2, size=(width, height, morph_count)),
            vetex_data_count,
            width,
            morph_count,
        )

    def get_pipeline_info(self, wobject, shared):
        material = wobject.material

        # The MeshMaterial can be applied to lines and points, so that we can fully support gltf
        if isinstance(wobject, Line):
            topology = wgpu.PrimitiveTopology.line_strip
            # todo: wgpu.PrimitiveTopology.line_list, wgpu.PrimitiveTopology.line_loop
        elif isinstance(wobject, Points):
            topology = wgpu.PrimitiveTopology.point_list
        elif isinstance(wobject, Mesh):
            topology = wgpu.PrimitiveTopology.triangle_list
            # todo: wgpu.PrimitiveTopology.triangle_strip
        else:
            raise TypeError(
                f"MeshMaterial cannot be applied to a {wobject.__class.__name__}"
            )

        if material.side == "front":
            cull_mode = wgpu.CullMode.back
        elif material.side == "back":
            cull_mode = wgpu.CullMode.front
        else:  # material.side == "both"
            cull_mode = wgpu.CullMode.none

        return {
            "primitive_topology": topology,
            "cull_mode": cull_mode,
        }

    def get_render_info(self, wobject, shared):
        geometry = wobject.geometry

        mul = self["indexer"]
        offset, size = geometry.indices.draw_range
        offset, size = mul * offset, mul * size

        n_instances = 1
        if self["instanced"]:
            n_instances = wobject.instance_buffer.nitems

        return {
            "indices": (size, n_instances, offset, 0),
        }

    def _check_texture(self, t, geometry, view_dim):
        assert isinstance(t, TextureMap)
        uv_channel = t.uv_channel
        if uv_channel > 0:
            texcoords = getattr(geometry, f"texcoords{uv_channel}", None)
        else:
            texcoords = getattr(geometry, "texcoords", None)
        assert texcoords is not None, (
            f"Texture {t} requires geometry.texcoords{uv_channel or ''}"
        )

        if view_dim == "1d":
            assert texcoords.data.ndim == 1 or texcoords.data.shape[-1] == 1, (
                f"Texture {t} requires 1D texcoords"
            )
        elif view_dim == "2d" or view_dim == "2d-array":
            assert texcoords.data.ndim == 2 and texcoords.data.shape[-1] == 2, (
                f"Texture {t} requires 2D texcoords"
            )
        elif view_dim == "cube" or view_dim == "3d" or view_dim == "cube-array":
            assert texcoords.data.ndim == 2 and texcoords.data.shape[-1] == 3, (
                f"Texture {t} requires 3D texcoords"
            )
        else:
            raise ValueError(f"Unknown view_dim: {view_dim}")

        fmt = to_texture_format(t.texture.format)
        assert "norm" in fmt or "float" in fmt

    def get_code(self):
        return load_wgsl("mesh.wgsl")


@register_wgpu_render_function(Mesh, MeshNormalMaterial)
class MeshNormalShader(MeshShader):
    def __init__(self, wobject):
        super().__init__(wobject)
        self["color_mode"] = "normal"
        self["colormap_dim"] = ""  # disable texture if there happens to be one


@register_wgpu_render_function(Mesh, MeshPhongMaterial)
class MeshPhongShader(MeshShader):
    def __init__(self, wobject):
        super().__init__(wobject)
        self["lighting"] = "phong"


@register_wgpu_render_function(Mesh, MeshToonMaterial)
class MeshToonShader(MeshShader):
    def __init__(self, wobject):
        super().__init__(wobject)
        self["lighting"] = "toon"

    def get_bindings(self, wobject, shared):
        result = super().get_bindings(wobject, shared)

        geometry = wobject.geometry
        material = wobject.material

        bindings = []

        if material.gradient_map is not None:
            bindings.extend(
                self._define_texture_map(
                    geometry, material.gradient_map, "gradient_map"
                )
            )
            self["use_gradient_map"] = True

        # Define shader code for binding
        bindings = {i: binding for i, binding in enumerate(bindings)}
        self.define_bindings(2, bindings)

        # Update result
        result[2] = bindings
        return result


@register_wgpu_render_function(WorldObject, MeshStandardMaterial)
class MeshStandardShader(MeshShader):
    def __init__(self, wobject):
        super().__init__(wobject)
        self["lighting"] = "pbr"

    def get_bindings(self, wobject, shared):
        result = super().get_bindings(wobject, shared)

        geometry = wobject.geometry
        material = wobject.material

        bindings = []

        if material.roughness_map is not None:
            bindings.extend(
                self._define_texture_map(
                    geometry, material.roughness_map, "roughness_map"
                )
            )
            self["use_roughness_map"] = True

        if material.metalness_map is not None:
            bindings.extend(
                self._define_texture_map(
                    geometry, material.metalness_map, "metalness_map"
                )
            )
            self["use_metalness_map"] = True

        # Define shader code for binding
        bindings = {i: binding for i, binding in enumerate(bindings)}
        self.define_bindings(2, bindings)

        # Update result
        result[2] = bindings
        return result


@register_wgpu_render_function(WorldObject, MeshPhysicalMaterial)
class MeshPhysicalShader(MeshStandardShader):
    def __init__(self, wobject):
        super().__init__(wobject)
        self["lighting"] = "pbr"
        self["USE_IOR"] = True
        self["USE_SPECULAR"] = True

    def get_bindings(self, wobject, shared):
        result = super().get_bindings(wobject, shared)

        geometry = wobject.geometry
        material = wobject.material

        bindings = []

        if material.specular_intensity_map is not None:
            bindings.extend(
                self._define_texture_map(
                    geometry, material.specular_intensity_map, "specular_intensity_map"
                )
            )
            self["use_specular_intensity_map"] = True

        # clearcoat
        if material.clearcoat:
            self["USE_CLEARCOAT"] = True

            if material.clearcoat_map is not None:
                bindings.extend(
                    self._define_texture_map(
                        geometry, material.clearcoat_map, "clearcoat_map"
                    )
                )
                self["use_clearcoat_map"] = True

            if material.clearcoat_roughness_map is not None:
                bindings.extend(
                    self._define_texture_map(
                        geometry,
                        material.clearcoat_roughness_map,
                        "clearcoat_roughness_map",
                    )
                )
                self["use_clearcoat_roughness_map"] = True

            if material.clearcoat_normal_map is not None:
                bindings.extend(
                    self._define_texture_map(
                        geometry, material.clearcoat_normal_map, "clearcoat_normal_map"
                    )
                )
                self["use_clearcoat_normal_map"] = True

        # iridescence
        if material.iridescence:
            self["USE_IRIDESCENCE"] = True

            if material.iridescence_map is not None:
                bindings.extend(
                    self._define_texture_map(
                        geometry, material.iridescence_map, "iridescence_map"
                    )
                )
                self["use_iridescence_map"] = True

            if material.iridescence_thickness_map is not None:
                bindings.extend(
                    self._define_texture_map(
                        geometry,
                        material.iridescence_thickness_map,
                        "iridescence_thickness_map",
                    )
                )
                self["use_iridescence_thickness_map"] = True

        # sheen
        if material.sheen:
            self["USE_SHEEN"] = True

            if material.sheen_color_map is not None:
                bindings.extend(
                    self._define_texture_map(
                        geometry, material.sheen_color_map, "sheen_color_map"
                    )
                )
                self["use_sheen_color_map"] = True

            if material.sheen_roughness_map is not None:
                bindings.extend(
                    self._define_texture_map(
                        geometry, material.sheen_roughness_map, "sheen_roughness_map"
                    )
                )
                self["use_sheen_roughness_map"] = True

        # anisotropy
        if material.anisotropy:
            self["USE_ANISOTROPY"] = True

            if material.anisotropy_map is not None:
                bindings.extend(
                    self._define_texture_map(
                        geometry, material.anisotropy_map, "anisotropy_map"
                    )
                )
                self["use_anisotropy_map"] = True

        # Define shader code for binding
        bindings = {i: binding for i, binding in enumerate(bindings)}
        self.define_bindings(3, bindings)

        # Update result
        result[3] = bindings
        return result


@register_wgpu_render_function(Mesh, MeshNormalLinesMaterial)
class MeshNormalLinesShader(MeshShader):
    def __init__(self, wobject):
        super().__init__(wobject)
        self["color_mode"] = "uniform"
        self["lighting"] = ""
        self["wireframe"] = False

    def get_pipeline_info(self, wobject, shared):
        d = super().get_pipeline_info(wobject, shared)
        d["primitive_topology"] = wgpu.PrimitiveTopology.line_list
        return d

    def get_render_info(self, wobject, shared):
        # We directly look at the vertex data, so geometry.indices.draw_range is ignored.
        d = super().get_render_info(wobject, shared)
        d["indices"] = wobject.geometry.positions.nitems * 2, d["indices"][1]
        return d

    def get_code(self):
        return load_wgsl("mesh_normal_lines.wgsl")


@register_wgpu_render_function(Mesh, MeshSliceMaterial)
class MeshSliceShader(BaseShader):
    """Shader for rendering mesh slices."""

    type = "render"

    def __init__(self, wobject):
        super().__init__(wobject)

        material = wobject.material
        geometry = wobject.geometry

        color_mode = str(material.color_mode).split(".")[-1]

        if color_mode == "auto":
            if material.map is not None:
                self["color_mode"] = "vertex_map"
                self["color_buffer_channels"] = 0
            else:
                self["color_mode"] = "uniform"
                self["color_buffer_channels"] = 0
        elif color_mode == "uniform":
            self["color_mode"] = "uniform"
            self["color_buffer_channels"] = 0
        elif color_mode == "vertex":
            nchannels = nchannels_from_format(geometry.colors.format)
            self["color_mode"] = "vertex"
            self["color_buffer_channels"] = nchannels
            if nchannels not in (1, 2, 3, 4):
                raise ValueError(f"Geometry.colors needs 1-4 columns, not {nchannels}")
        elif color_mode == "face":
            nchannels = nchannels_from_format(geometry.colors.format)
            self["color_mode"] = "face"
            self["color_buffer_channels"] = nchannels
            if nchannels not in (1, 2, 3, 4):
                raise ValueError(f"Geometry.colors needs 1-4 columns, not {nchannels}")
        elif color_mode == "vertex_map":
            self["color_mode"] = "vertex_map"
            self["color_buffer_channels"] = 0
            if material.map is None:
                raise ValueError("Cannot apply colormap is no material.map is set.")
        elif color_mode == "face_map":
            self["color_mode"] = "face_map"
            self["color_buffer_channels"] = 0
            if material.map is None:
                raise ValueError("Cannot apply colormap is no material.map is set.")
        else:
            raise RuntimeError(f"Unknown color_mode: '{color_mode}'")

    def get_bindings(self, wobject, shared):
        # It would technically be possible to implement colormapping or
        # per-vertex colors, but its a tricky dance to get the per-vertex
        # data (e.g. texcoords) into a varying. And because the visual
        # result is a line, its likely that in most use-cases a uniform
        # color is preferred anyway. So for now we don't implement that.

        geometry = wobject.geometry
        material = wobject.material

        # Init uniform bindings
        bindings = [
            Binding("u_stdinfo", "buffer/uniform", shared.uniform_buffer),
            Binding("u_wobject", "buffer/uniform", wobject.uniform_buffer),
            Binding("u_material", "buffer/uniform", material.uniform_buffer),
        ]

        # We're assuming the presence of an index buffer for now
        assert getattr(geometry, "indices", None)

        # Init storage buffer bindings
        rbuffer = "buffer/read_only_storage"
        bindings.append(Binding("s_indices", rbuffer, geometry.indices, "VERTEX"))
        bindings.append(Binding("s_positions", rbuffer, geometry.positions, "VERTEX"))

        # Bindings for color
        if self["color_mode"] in ("vertex", "face"):
            bindings.append(Binding("s_colors", rbuffer, geometry.colors, "VERTEX"))
        elif self["color_mode"] in ("vertex_map", "face_map"):
            bindings.extend(
                self.define_generic_colormap(material.map, geometry.texcoords)
            )

        # Let the shader generate code for our bindings
        bindings = {i: binding for i, binding in enumerate(bindings)}
        self.define_bindings(0, bindings)

        return {
            0: bindings,
        }

    def get_pipeline_info(self, wobject, shared):
        return {
            "primitive_topology": wgpu.PrimitiveTopology.triangle_list,
            "cull_mode": wgpu.CullMode.none,
        }

    def get_render_info(self, wobject, shared):
        geometry = wobject.geometry
        offset, size = geometry.indices.draw_range
        offset, size = offset * 6, size * 6
        return {
            "indices": (size, 1, offset, 0),
        }

    def get_code(self):
        return load_wgsl("mesh_slice.wgsl")



"""
Detect fonts on the system. This code was inspired by the Matplotlib
font manager. It differs in that we rely fully on finding fonts in
certain directories. We do not use the Windows registry or anything
like that to get a list of "official" fonts. The downside is that we
may find more fonts that the OS feels are registered. The big advantage
is that we can check the mtime of a handful of directories to know
whether we need to update our cache (and can also do partial updates).
"""

import os
import sys
import json
import time
import secrets

import freetype

from .. import logger, get_resources_dir, get_cache_dir


# Weight names according to CSS and the OpenType spec.
weight_dict = {
    "thin": 100,
    "hairline": 100,
    "ultralight": 200,
    "extralight": 200,
    "light": 300,
    "normal": 400,
    "regular": 400,
    "medium": 500,
    "semibold": 600,
    "demibold": 600,
    "bold": 700,
    "extrabold": 800,
    "ultrabold": 800,
    "black": 900,
    "heavy": 900,
}


style_dict = {
    "normal": "normal",
    "regular": "normal",
    "italic": "italic",
    "oblique": "oblique",
}


class FontFile:
    """Object to represent a font file."""

    def __init__(self, filename, family=None, variant=None, codepoints=None):
        assert isinstance(filename, str)
        assert family is None or isinstance(family, str)
        assert variant is None or isinstance(variant, str)
        assert codepoints is None or isinstance(codepoints, set)

        self._filename = filename
        self._family = family
        self._variant = variant
        self._name = None
        self._weight = None
        self._style = None
        self._codepoints = codepoints

    def __repr__(self):
        return f"<FontFile {self.name} at 0x{hex(id(self))}>"

    def __hash__(self):
        return hash(self.name)

    def _get_face(self):
        # This was factored out so it can be overloaded in tests
        return freetype.Face(self._filename)

    @property
    def filename(self):
        """The path to this font file."""
        return self._filename

    @property
    def family(self):
        """The family name of this font, e.g. 'Noto Sans' or 'Arial'.
        This value is defined in the font file.
        """
        if not self._family:
            self._family = self._get_face().family_name.decode()
            if not self._family:
                name = os.path.basename(self._filename).split(".")[0]
                family, _, _ = name.partition("-")
                self._family = family or "Unknown"
        return self._family

    @property
    def variant(self):
        """The variant name of this font, e.g. 'Regular', 'Bold',
        'Italic', 'Thin Italic'. This is a value defined in the font
        file, and from this the weight and style are derived.
        """
        if not self._variant:
            self._variant = self._get_face().style_name.decode()
            if not self._variant:
                name = os.path.basename(self._filename).split(".")[0]
                _, _, variant = name.partition("-")
                self._variant = variant or "Regular"
        return self._variant

    @property
    def weight(self):
        """The font weight, as a number between 100-900."""
        if not self._weight:
            variant_name = self._variant.lower()
            if variant_name == "regular":  # make common cases fast
                self._weight = 400
            else:
                for weight_name, weight_nr in weight_dict.items():
                    if weight_name in variant_name:
                        self._weight = weight_nr
                        break
                else:
                    self._weight = 400
        return self._weight

    @property
    def style(self):
        """The style, one of "normal", "italic", or "oblique"."""
        if not self._style:
            variant_name = self._variant.lower()
            for style_name1, style_name2 in style_dict.items():
                if style_name1 in variant_name:
                    self._style = style_name2
                    break
            else:
                self._style = "normal"
        return self._style

    @property
    def name(self):
        """A normalized name that includes the family and variant. This
        therefore uniquely identifies the font. This typically is the
        same as the filename (without extension), but this is not
        guaranteed to be the case.
        """
        if not self._name:
            family = "".join(x[0].upper() + x[1:] for x in self.family.split())
            style = "".join(x[0].upper() + x[1:] for x in self.variant.split())
            self._name = family + "-" + style
        return self._name

    @property
    def codepoints(self):
        """A set of Unicode code points (ints) supported by this font.
        To test whether a certain codepoint is supported, use has_codepoint() instead.
        """
        # note: we could use a data structure that stores codepoints more efficiently
        if self._codepoints is None:
            self._codepoints = set(i for i, _ in self._get_face().get_chars())
        return self._codepoints

    def has_codepoint(self, codepoint):
        """Check whether a codepoint is supported by this font."""
        return codepoint in self.codepoints


def get_all_fonts():
    """Get a set of all available fonts."""
    # Disabling system fonts can be very helpful in testing to
    # ensure that we get repeatable results nomatter the version of the
    # fonts installed on the machine.
    disable_system_fonts = os.environ.get("PYGFX_DISABLE_SYSTEM_FONTS", "0").lower()
    if disable_system_fonts in ("1", "true", "yes"):
        return get_builtin_fonts()
    else:
        return get_builtin_fonts() | get_system_fonts()


def get_builtin_fonts():
    """Get a list of fonts that are shipped with pygfx."""
    _dir_paths, file_paths = find_fonts_paths(get_resources_dir(), False)
    return {FontFile(p) for p in file_paths}


def get_entrypoints_fonts():
    """Get a list of fonts defined via the setuptools entrypoint. TODO someday."""
    raise NotImplementedError()


def get_system_fonts():
    """Get a list of system fonts. The list is cached for speed. When
    this function is called, it checks whether any of the directories
    that may contain fonts have changed (files are deleted, added, or
    renamed). If so, we search that directory again. This way we are
    able to detect new fonts without an exhaustive search each process
    startup (which would be slow).
    """

    # Load cache from file system
    filename = os.path.join(get_cache_dir(), "font_cache.json")
    try:
        with open(filename, "rt", encoding="utf-8") as f:
            cache = json.load(f)
    except (FileNotFoundError, json.JSONDecodeError):
        cache = {}

    reset = False
    dirs_to_update = set()
    new_dir_paths = set()
    new_file_paths = set()

    # Check if any of the directories have updated. Also basic sanity checks.
    if cache:
        try:
            for p, info in cache["dirs"].items():
                if not isinstance(p, str):
                    raise RuntimeError("Dir path not a str.")
                elif not isinstance(info, dict):
                    raise RuntimeError("Dir info not a dict.")
                elif not os.path.isdir(p):
                    reset = True
                elif info["mtime"] < os.path.getmtime(p):
                    dirs_to_update.add(p)
            for p, info in cache["files"].items():
                if not isinstance(p, str):
                    raise RuntimeError("File path not a str.")
                elif not isinstance(info, dict):
                    raise RuntimeError("File info not a dict.")
                elif {"mtime", "family", "variant"}.difference(info.keys()):
                    reset = True  # We probably added new info to the cache
                # elif not os.path.isfile(p):  # slow, and should be covered by dir mtime
                #     dirs_to_update(os.path.dirname(p))
        except Exception as err:
            logger.error("Error loading font cache: %s", err)
            cache = {}

    # Reset directories that need an update
    for dir_path in dirs_to_update:
        dir_path = os.path.normpath(dir_path)
        # Clear fonts in this dir
        files = cache["files"]
        for p in list(files.keys()):
            if os.path.normpath(os.path.dirname(p)) == dir_path:
                files.pop(p)
        # Detect fonts again, schedule for inclusion in the cache
        _, file_paths = find_fonts_paths(dir_path, False)
        new_dir_paths.add(dir_path)
        new_file_paths.update(file_paths)

    # Do a full search if needed
    if reset or not cache:
        new_dir_paths, new_file_paths = find_system_fonts()
        cache = {"dirs": {}, "files": {}}

    # Do we need to update?
    if new_dir_paths or new_file_paths:
        logger.info(f"Searched for fonts in: {new_dir_paths}")

        # Put new dirs in the cache
        dirs = cache["dirs"]
        for p in new_dir_paths:
            dirs[p.replace("\\", "/")] = {
                "mtime": os.path.getmtime(p),
            }

        # Put new files in the cache
        files = cache["files"]
        for p in new_file_paths:
            ff = FontFile(p)
            try:
                ff.name  # noqa: B018 - This makes FreeType open the file
            except Exception:
                continue
            files[p.replace("\\", "/")] = {
                "mtime": os.path.getmtime(ff.filename),
                "family": ff.family,
                "variant": ff.variant,
            }

        # Sort the keys, just being clean
        cache = {"dirs": {}, "files": {}}
        for key in sorted(dirs.keys()):
            cache["dirs"][key] = dirs[key]
        for key in sorted(files.keys()):
            cache["files"][key] = files[key]

        # Write to file system, first to a uniquely named file
        filename2 = filename + ".part." + secrets.token_urlsafe(4)
        with open(filename2, "wt", encoding="utf-8") as f:
            json.dump(cache, f)

        # Now we must replace the font cache on disk with the result.
        # But we must take into account that another process might be
        # reading from it. The os.replace() call is atomic. And on Unix
        # that's that - whatever is reading is can keep doing so,
        # because this is simply a new inode with the same name. On
        # Windows, however, the replace will fail if something else has
        # the file open. So we wait and try again, until we time out
        # and give up.
        etime = time.time() + 5
        while time.time() < etime:
            try:
                os.replace(filename2, filename)
            except OSError:
                time.sleep(0.1)
            else:
                break
        else:
            logger.error("Failed to write font cache")
            try:
                os.remove(filename2)
            except Exception:
                pass

    # Return set of FontFile objects
    return {
        FontFile(filename, info["family"], info["variant"])
        for filename, info in cache["files"].items()
    }


def find_system_fonts():
    """Find fonts on the file system. Returns (dir_paths, file_paths).
    This function might be relatively slow.
    """
    dirs = get_system_font_directories()
    dir_paths = set()
    file_paths = set()
    for d in dirs:
        new_dirs, new_files = find_fonts_paths(d, True)
        dir_paths.update(new_dirs)
        file_paths.update(new_files)
    return dir_paths, file_paths


def find_fonts_paths(directory, recursive):
    """Return two sets (dir_paths, file_paths) representing the dirs and
    fonts matching any of the extensions, found in the given directory.
    """
    if not os.path.isdir(directory):
        raise OSError(f"Not a directory: {directory}")
    extensions = ".ttf", ".otf"
    dir_paths = set()
    file_paths = set()
    if not recursive:
        dir_paths.add(directory)
        for fname in os.listdir(directory):
            if fname.lower().endswith(extensions):
                file_paths.add(os.path.join(directory, fname))
    else:
        for dirpath, _, filenames in os.walk(directory):
            dir_paths.add(dirpath)
            for fname in filenames:
                if fname.lower().endswith(extensions):
                    file_paths.add(os.path.join(dirpath, fname))
    return dir_paths, file_paths


# %% OS-specific logic


def get_system_font_directories():
    """Get a set of system font directories."""
    # Simple triage, easy to replace in tests.
    if sys.platform.startswith("win"):
        return get_windows_font_directories()
    elif sys.platform.startswith("darwin"):
        return get_osx_font_directories()
    else:
        return get_unix_font_directories()


def get_windows_font_directories():
    import winreg

    dirs = set()
    dirs.update(WinFontDirs)
    dirs.add(os.path.join(os.getenv("WINDIR", ""), "Fonts"))

    # Get win32 font directory from the registry
    ms_folders = r"Software\Microsoft\Windows\CurrentVersion\Explorer\Shell Folders"
    try:
        with winreg.OpenKey(winreg.HKEY_CURRENT_USER, ms_folders) as user:
            dirs.add(winreg.QueryValueEx(user, "Fonts")[0])
    except OSError:
        pass

    return {os.path.abspath(d) for d in dirs if os.path.isdir(d)}


def get_osx_font_directories():
    dirs = set()
    dirs.update(X11FontDirectories)
    dirs.update(OSXFontDirectories)
    return {os.path.abspath(d) for d in dirs if os.path.isdir(d)}


def get_unix_font_directories():
    # MPL also calls out to fontconfig (fc-list) to get a list of fonts.
    # But if we assume that our list of possible directories is
    # complete, this should not be necessary. Let's see how it goes.
    dirs = set()
    dirs.update(X11FontDirectories)
    return {os.path.abspath(d) for d in dirs if os.path.isdir(d)}


try:
    HOME = os.path.expanduser("~")
except Exception:  # Exceptions thrown by home() are not specified...
    HOME = "/home"  # Just an arbitrary path

WinFontDirs = [
    os.path.join(os.getenv("LOCALAPPDATA", ""), "Microsoft/Windows/Fonts"),
    os.path.join(os.getenv("APPDATA", ""), "Microsoft/Windows/Fonts"),
    os.path.join(
        os.getenv("ALLUSERSPROFILE", ""), "AppData/Local/Microsoft/Windows/Fonts"
    ),
    os.path.join(
        os.getenv("ALLUSERSPROFILE", ""), "AppData/Roaming/Microsoft/Windows/Fonts"
    ),
]

X11FontDirectories = [
    # an old standard installation point
    "/usr/X11R6/lib/X11/fonts/TTF/",
    "/usr/X11/lib/X11/fonts",
    # here is the new standard location for fonts
    "/usr/share/fonts/",
    # documented as a good place to install new fonts
    "/usr/local/share/fonts/",
    # common application, not really useful
    "/usr/lib/openoffice/share/fonts/truetype/",
    # user fonts
    os.path.join(os.getenv("XDG_DATA_HOME") or HOME, ".local/share/fonts"),
    os.path.join(HOME, ".fonts"),
]

OSXFontDirectories = [
    "/Library/Fonts/",
    "/Network/Library/Fonts/",
    "/System/Library/Fonts/",
    # fonts installed via MacPorts
    "/opt/local/share/fonts",
    # user fonts
    os.path.join(HOME, "Library/Fonts"),
]



"""
Eye-Dome Lighting (EDL) post-processing pass for pygfx.

This pass is inspired by Potree's EDL implementation and
applies a depth-based shading in screen space to enhance the perception of
surfaces in dense point clouds or meshes.

References:
- Potree (EDL shader and technique): https://github.com/potree/potree
"""

from .effectpasses import EffectPass


class EDLPass(EffectPass):
    """Eye-Dome Lighting post-effect (Potree-like).

    Parameters
    ----------
    strength : float
        EDL strength. Typical range ~ [0.5, 10.0]. Default 1.0.
    radius : float
        Sampling radius in pixels. Typical range ~ [1.0, 3.0]. Default 1.5.
    num_samples : int
        Number of neighbor samples around the pixel. Typical values 4 or 8.
    depth_edge_threshold : float
        Threshold to gate the response for very small depth differences.
    """

    USES_DEPTH = True

    uniform_type = dict(
        EffectPass.uniform_type,
        strength="f4",
        radius="f4",
        depth_edge_threshold="f4",
    )

    wgsl = """
        const EDL_SCALE: f32 = 200.0; // scale to match Potree-like response levels

        @fragment
        fn fs_main(varyings: Varyings) -> @location(0) vec4<f32> {
            let texIndex = vec2i(varyings.position.xy);

            // Load center depth and color
            let depth_c = textureLoad(depthTex, texIndex, 0);
            let color_c = textureLoad(colorTex, texIndex, 0);

            // Early out if depth is invalid (background)
            if (depth_c >= 1.0) {
                return color_c;
            }

            // Neighbor sampling pattern (8 directions). We sample two rings: r and 2r (total 16 taps)
            // Currently shader is fixed at 8 taps; future: template var for 4 taps
            let dir = array<vec2<f32>, 8>(
                vec2<f32>(1.0, 0.0),
                vec2<f32>(0.70710678, 0.70710678),
                vec2<f32>(0.0, 1.0),
                vec2<f32>(-0.70710678, 0.70710678),
                vec2<f32>(-1.0, 0.0),
                vec2<f32>(-0.70710678, -0.70710678),
                vec2<f32>(0.0, -1.0),
                vec2<f32>(0.70710678, -0.70710678)
            );

            let dims = textureDimensions(depthTex);
            let max_i = vec2<i32>(i32(dims.x) - 1, i32(dims.y) - 1);

            var response: f32 = 0.0;
            // Ring 1 (radius r)
            for (var i: i32 = 0; i < 8; i = i + 1) {
                let off = vec2<i32>(
                    i32(round(dir[i].x * u_effect.radius)),
                    i32(round(dir[i].y * u_effect.radius)),
                );
                var ni = texIndex + off;
                ni = clamp(ni, vec2<i32>(0, 0), max_i);
                let depth_n = textureLoad(depthTex, ni, 0);
                let is_bg = depth_n >= 0.9999;
                if (!is_bg) {
                    // Positive when neighbor is further (darker)
                    let dd = depth_n - depth_c;
                    let contrib = max(0.0, dd - u_effect.depth_edge_threshold);
                    response = response + contrib;
                }
            }
            // Ring 2 (radius 2r), weighted lower
            for (var i: i32 = 0; i < 8; i = i + 1) {
                let off = vec2<i32>(
                    i32(round(dir[i].x * (u_effect.radius * 2.0))),
                    i32(round(dir[i].y * (u_effect.radius * 2.0))),
                );
                var ni = texIndex + off;
                ni = clamp(ni, vec2<i32>(0, 0), max_i);
                let depth_n = textureLoad(depthTex, ni, 0);
                let is_bg = depth_n >= 0.9999;
                if (!is_bg) {
                    let dd = depth_n - depth_c;
                    let contrib = max(0.0, dd - u_effect.depth_edge_threshold) * 0.5;
                    response = response + contrib;
                }
            }

            // Normalize and map response to a dimming factor
            // scale with EDL_SCALE for visible effect
            let shade = exp(-u_effect.strength * response * EDL_SCALE);
            let shaded_rgb = color_c.rgb * shade;
            return vec4<f32>(shaded_rgb, color_c.a);
        }
    """

    def __init__(self, *, strength=1.0, radius=1.5, depth_edge_threshold=0.0):
        super().__init__()
        self.strength = strength
        self.radius = radius
        self.depth_edge_threshold = depth_edge_threshold

    # Properties
    @property
    def strength(self):
        return float(self._uniform_data["strength"])

    @strength.setter
    def strength(self, value):
        self._uniform_data["strength"] = float(value)

    @property
    def radius(self):
        return float(self._uniform_data["radius"])

    @radius.setter
    def radius(self, value):
        self._uniform_data["radius"] = float(value)

    @property
    def depth_edge_threshold(self):
        return float(self._uniform_data["depth_edge_threshold"])

    @depth_edge_threshold.setter
    def depth_edge_threshold(self, value):
        self._uniform_data["depth_edge_threshold"] = float(value)



from typing import Tuple

import numpy as np
import pylinalg as la

from ._panzoom import PanZoomController


def _get_axis_aligned_up_vector(up):
    # Not actually used, but I have a feeling we might need it at some point :)
    ref_up, largest_dot = None, 0
    for up_vec in [(1, 0, 0), (0, 1, 0), (0, 0, 1), (-1, 0, 0), (0, -1, 0), (0, 0, -1)]:
        up_vec = np.array(up_vec)
        v = np.dot(up_vec, up)
        if v > largest_dot:
            ref_up, largest_dot = up_vec, v
    return ref_up


class OrbitController(PanZoomController):
    """A controller to move a camera in an orbit around a center position.

    Supports panning parallel to the screen, zooming, orbiting.

    Parameters
    ----------
    target : tuple of float, optional
        The custom target position (x, y, z) that the camera orbits around.
        Default is None, which means the target is determined from the camera state.

    Notes
    -----
    The direction of rotation is defined such that it feels like you're
    grabbing onto something in the foreground; if you move the mouse
    to the right, the objects in the foreground move to the right, and
    those in the background (on the opposite side of the center of rotation)
    move to the left.

    Default controls:

    * Left mouse button: orbit / rotate.
    * Right mouse button: pan.
    * Fourth mouse button: quickzoom
    * wheel: zoom to point.
    * alt+wheel: adjust fov.

    """

    _default_controls = {
        "mouse1": ("rotate", "drag", (0.005, 0.005)),
        "mouse2": ("pan", "drag", (1, 1)),
        "mouse4": ("quickzoom", "peek", 2),
        "wheel": ("zoom", "push", -0.001),
        "alt+wheel": ("fov", "push", -0.01),
    }

    def __init__(self, *args, target=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.target = np.array(target) if target is not None else None

    @property
    def target(self):
        """The target position (x, y, z) that the camera orbits around.

        Set to None to use an implicit target based on the camera state. This only works
        well in combination with ``camera.show_object()`` and ``camera.show_pos()``.
        """
        if self._custom_target is not None:
            return self._custom_target.copy()
        else:
            camera_state = self._get_camera_state()
            return self._get_target_vec(camera_state)

    @target.setter
    def target(self, value):
        if value is None:
            self._custom_target = None
        else:
            self._custom_target = np.array(value, dtype=np.float32)
            for camera, _, _ in self._cameras:
                camera.look_at(self._custom_target)

    def rotate(self, delta: Tuple, rect: Tuple, *, animate=False):
        """Rotate in an orbit around the target, using two angles (azimuth and elevation, in radians).

        If animate is True, the motion is damped. This requires the
        controller to receive events from the renderer/viewport.
        """

        if animate:
            action_tuple = ("rotate", "push", (1.0, 1.0))
            action = self._create_action(None, action_tuple, (0.0, 0.0), None, rect)
            action.set_target(delta)
            action.snap_distance = 0.01
            action.done = True
        elif self._cameras:
            self._update_rotate(delta)
            return self._update_cameras()

    def _update_rotate(self, delta):
        assert isinstance(delta, tuple) and len(delta) == 2

        delta_azimuth, delta_elevation = delta
        camera_state = self._get_camera_state()

        # Note: this code does not use la.vec_euclidean_to_spherical and
        # la.vec_spherical_to_euclidean, because those functions currently
        # have no way to specify a different up vector.

        position = camera_state["position"]
        rotation = camera_state["rotation"]
        up = camera_state["reference_up"]

        # Where is the camera looking at right now
        forward = la.vec_transform_quat((0, 0, -1), rotation)

        # # Get a reference vector, that is orthogonal to up, in a deterministic way.
        # # Might need this if we ever want the azimuth
        # aligned_up = _get_axis_aligned_up_vector(up)
        # orthogonal_vec = np.cross(up, np.roll(aligned_up, 1))

        # Get current elevation, so we can clip it.
        # We don't need the azimuth. When we do, it'd need more care to get a proper 0..2pi range
        elevation = la.vec_angle(forward, up) - 0.5 * np.pi

        # Apply boundaries to the elevation
        new_elevation = elevation + delta_elevation
        bounds = -89 * np.pi / 180, 89 * np.pi / 180
        if new_elevation < bounds[0]:
            delta_elevation = bounds[0] - elevation
        elif new_elevation > bounds[1]:
            delta_elevation = bounds[1] - elevation

        r_azimuth = la.quat_from_axis_angle(up, -delta_azimuth)
        r_elevation = la.quat_from_axis_angle((1, 0, 0), -delta_elevation)

        # Get rotations
        rot1 = rotation
        rot2 = la.quat_mul(r_azimuth, la.quat_mul(rot1, r_elevation))

        # Calculate new position
        pos1 = position
        if self._custom_target is not None:
            # If we have a custom target, we need to calculate the target vector

            target_pos = self._custom_target
            pos1_to_target = target_pos - pos1

            pos1_to_target_rotated = la.vec_transform_quat(pos1_to_target, r_azimuth)

            right = la.vec_transform_quat((1, 0, 0), rot1)
            r_elevation_world = la.quat_from_axis_angle(right, -delta_elevation)
            pos1_to_target_final = la.vec_transform_quat(
                pos1_to_target_rotated, r_elevation_world
            )

            pos2 = target_pos - pos1_to_target_final
        else:
            pos2target1 = self._get_target_vec(camera_state, rotation=rot1)
            pos2target2 = self._get_target_vec(camera_state, rotation=rot2)
            pos2 = pos1 + pos2target1 - pos2target2

        # Apply new state
        new_camera_state = {"position": pos2, "rotation": rot2}
        self._set_camera_state(new_camera_state)

        # Note that for ortho cameras, we also orbit around the scene,
        # even though it could be positioned at the center (i.e.
        # target). Doing it this way makes the code in the controllers
        # easier. The only downside I can think of is that the far plane
        # is now less far away but this effect is only 0.2%, since the
        # far plane is 500 * dist.

    def _update_pan(self, delta, *, vecx, vecy):
        if self._custom_target is not None:
            # If we have a custom target, we need to pan relative to that
            target_pos = self._custom_target

            distance_to_target = la.vec_dist(
                target_pos, self._get_camera_state()["position"]
            )

            scaled_delta = (
                delta[0] * distance_to_target * 0.01,
                delta[1] * distance_to_target * 0.01,
            )
            offest = -vecx * scaled_delta[0] + vecy * scaled_delta[1]
            self._custom_target = target_pos + offest
        else:
            scaled_delta = delta

        return super()._update_pan(scaled_delta, vecx=vecx, vecy=vecy)

    def _update_zoom(self, delta):
        if isinstance(delta, (int, float)):
            delta = (delta, delta)
        assert isinstance(delta, tuple) and len(delta) == 2

        if self._custom_target is not None:
            target_pos = self._custom_target
            radius_vec = self._get_camera_state()["position"] - target_pos

            scale = 1 - delta[0]
            if scale < 0.1:
                scale = 0.1

            radius_vec *= scale
            new_position = target_pos + radius_vec
            self._set_camera_state(
                {
                    "position": new_position,
                }
            )

        else:
            fx = 2 ** delta[0]
            fy = 2 ** delta[1]
            new_cam_state = self._zoom(fx, fy, self._get_camera_state())
            self._set_camera_state(new_cam_state)



from __future__ import annotations

import random
import weakref
import threading
from typing import Any, Callable, ClassVar, Iterator, List, Tuple
import pylinalg as la
from time import perf_counter_ns

import numpy as np

from ..resources import Buffer, Texture
from ..utils import array_from_shadertype, logger
from ..utils.trackable import Trackable
from ..utils.bounds import Bounds
from ._events import EventTarget
from ..utils.transform import (
    AffineTransform,
    RecursiveTransform,
)
from ..geometries import Geometry
from ..materials import Material


class IdProvider:
    """Object for internal use to manage world object id's."""

    def __init__(self):
        self._ids_in_use = set([0])
        self._map = weakref.WeakValueDictionary()
        self._lock = threading.RLock()

    def claim_id(self, wobject: WorldObject) -> int:
        """Used by wobjects to claim an id."""
        # We don't simply count up, but keep a pool of ids. This is
        # because an application *could* create and discard objects at
        # a high rate, so we want to be able to re-use these ids.
        #
        # Some numbers:
        # * 4_294_967_296 (2**32) max number for u32
        # * 2_147_483_647 (2**31 -1) max number for i32.
        # *    16_777_216 max integer that can be stored exactly in f32
        # *     4_000_000 max integer that survives being passed as a varying (in my tests)
        # *     1_048_575 is ~1M is 2**20 seems like a good max scene objects.
        # *    67_108_864 is ~50M is 2**26 seems like a good max vertex count.
        #                 which leaves 64-20-26=18 bits for any other picking info.

        # Max allowed id, inclusive
        id_max = 1_048_575  # 2*20-1

        # The max number of ids. This is a bit less to avoid choking
        # when there are few free id's left.
        max_items = 1_000_000

        with self._lock:
            if len(self._ids_in_use) >= max_items:
                raise RuntimeError("Max number of objects reached.")
            id = 0
            while id in self._ids_in_use:
                id = random.randint(1, id_max)
            self._ids_in_use.add(id)
            self._map[id] = wobject

        return id

    def release_id(self, wobject: WorldObject, id: int) -> None:
        """Release an id associated with a wobject."""
        if id > 0:
            with self._lock:
                self._ids_in_use.discard(id)
                self._map.pop(id, None)

    def get_object_from_id(self, id: int) -> WorldObject | None:
        """Return the wobject associated with an id, or None."""
        return self._map.get(id)


id_provider = IdProvider()


class WorldObject(EventTarget, Trackable):
    """Base class for objects.

    This class represents objects in the world, i.e., the scene graph.Each
    WorldObject has geometry to define it's data, and material to define its
    appearance. The object itself is only responsible for defining object
    hierarchies (parent / children) and its position and orientation in the
    world.

    Parameters
    ----------
    geometry : Geometry
        The data defining the shape of the object. See the documentation
        on the different WorldObject subclasses for what attributes the
        geometry should and may have.
    material : Material
        The data defining the appearance of the object.
    visible : bool
        Whether the object is visible.
    render_order : float
        Value that helps controls the order in which objects are rendered.
    name : str
        The name of the object.

    Notes
    -----
    Use :class:`Group` to collect multiple world objects into a single empty
    world object.

    See Also
    --------
    pygfx.utils.transform.AffineBase
        Various getters and setters defined on ``obj.local`` and ``obj.world``.
    pygfx.utils.transform.AffineTransform
        The class used to implement ``obj.local``.
    pygfx.utils.transform.RecursiveTransform
        The class used to implement ``obj.world``.

    """

    _FORWARD_IS_MINUS_Z = False  # Default is +Z (lights and cameras use -Z)

    _id = 0

    # The uniform type describes the structured info for this object, which represents
    # every "property" that a renderer would need to know in order to visualize it.
    # Put larger items first for alignment, also note that host-sharable structs
    # align at power-of-two only, so e.g. vec3 needs padding.
    # todo: rename uniform to info or something?

    uniform_type: ClassVar[dict[str, str]] = dict(
        world_transform="4x4xf4",
        world_transform_inv="4x4xf4",
        global_id="u4",
        renderer_id="u4",
    )

    def __init__(
        self,
        geometry: Geometry | None = None,
        material: Material | None = None,
        *,
        visible: bool = True,
        render_order: float = 0,
        name: str = "",
    ) -> None:
        super().__init__()
        self._parent: weakref.ReferenceType[WorldObject] | None = None

        #: Subtrees of the scene graph that depend on this object.
        self._children: List[WorldObject] = []

        self.geometry = geometry
        self.material = material

        self.name = name

        # Compose complete uniform type
        buffer = Buffer(array_from_shadertype(self.uniform_type), force_contiguous=True)
        buffer.data["world_transform"] = np.eye(4)
        buffer.data["world_transform_inv"] = np.eye(4)

        self._world_last_modified = perf_counter_ns()

        #: The object's transform expressed in parent space.
        self.local = AffineTransform(is_camera_space=self._FORWARD_IS_MINUS_Z)
        #: The object's transform expressed in world space.
        self.world = RecursiveTransform(
            self.local, is_camera_space=self._FORWARD_IS_MINUS_Z, reference_up=(0, 1, 0)
        )

        # Set id that is global to the process. For identification of the object
        self._global_id = id_provider.claim_id(self)
        buffer.data["global_id"] = self._global_id

        # Init id assigned by the renderer. This id will be set once, by the first
        # renderer to render this object. That way, within one "visualization"
        # (consisting of a renderer and a scene with fresh objects) these id's are
        # reproducable. I.e. validation examples have reproducable locally unique id's.
        self._renderer_id = 0

        # Bounds
        self._bounds_geometry = None
        self._bounds_geometry_rev = 0

        #: The GPU data of this WorldObject.
        self.uniform_buffer = buffer

        # Init visibility and render props
        self.visible = visible
        self.render_order = render_order
        self.cast_shadow = False
        self.receive_shadow = False

        self.name = name

    def _assign_renderer_id(self, id):
        if self._renderer_id == 0:
            assert id > 0
            self._renderer_id = id
            self.uniform_buffer.data["renderer_id"] = id
            self.uniform_buffer.update_full()
            return 1
        return 0

    def _update_object(self):
        """This gets called (by the renderer) right before being drawn. Good time for lazy updates."""
        world_last_modified = self.world.last_modified
        if world_last_modified > self._world_last_modified:
            self._world_last_modified = world_last_modified
            self._update_world_transform()

    def _update_world_transform(self):
        """This gets called right before being drawn, when the world transform has changed."""
        orig_err_setting = np.seterr(under="ignore")
        self.uniform_buffer.data["world_transform"] = self.world.matrix.T
        self.uniform_buffer.data["world_transform_inv"] = self.world.inverse_matrix.T
        self.uniform_buffer.update_full()
        np.seterr(**orig_err_setting)

    def __repr__(self):
        return f"<pygfx.{self.__class__.__name__} {self.name} at {hex(id(self))}>"

    def __del__(self):
        id_provider.release_id(self, self.id)
        self.local._set_wrapper(
            None
        )  # break the circular reference so GC has it a little easier

    def _self(self):
        """Get self. This looks silly, but in case the WorldObject is
        wrapped in a weakproxy, like FastPlotLib does, this gets the
        real wobject, so we can hash it. You cannot dereference a weak.proxy.
        """
        return self

    @property
    def up(self) -> np.ndarray:
        """
        Relic of old WorldObjects that aliases with the new ``transform.up``
        direction. Prefer `obj.world.reference_up` instead.

        """

        logger.warning(
            "`WorldObject.up` is deprecated. Use `WorldObject.world.reference_up` instead.",
        )

        return self.world.reference_up

    @up.setter
    def up(self, value: np.ndarray) -> None:
        logger.warning(
            "`WorldObject.up` is deprecated. Use `WorldObject.world.reference_up` instead.",
        )

        self.world.reference_up = np.asarray(value)

    @property
    def id(self) -> int:
        """An integer id smaller than 2**31 (read-only)."""
        return self._global_id

    @property
    def visible(self) -> bool:
        """Whether is object is rendered or not. Default True."""
        return self._store.visible

    @visible.setter
    def visible(self, visible: bool) -> None:
        self._store.visible = bool(visible)

    @property
    def render_order(self) -> float:
        """Per-object rendering priority used to fine-tune the draw order within a render queue.

        Objects with higher render_order values are rendered later than those with lower values.
        This affects both opaque and transparent objects and can be used to resolve z-fighting,
        or control draw order beyond automatic depth sorting.

        The effective render order is the sum of its render order and thet of all its parents.

        The order in wich objects are rendered is:
            1. the ``material.render_queue``.
            2. the effective ``render_order``.
            3. the distance to camera (if ``render.sort_objects==True``).
            4. the position of the object in the scene graph.

        Also see ``material.render_queue``.
        """
        # Note: the render order is on the object, not the material, because it affects
        # a specific object, and materials are often shared between multiple objects.
        return self._store.render_order

    @render_order.setter
    def render_order(self, value: float) -> None:
        self._store.render_order = float(value)

    @property
    def render_mask(self):
        return None

    @render_mask.setter
    def render_mask(self, value):
        raise DeprecationWarning(
            "render_mask is deprecated, see material.alpha_mode to control how the renderer should treat an object."
        )

    @property
    def geometry(self) -> Geometry | None:
        """The object's geometry, the data that defines (the shape of) this object."""
        return self._store.geometry

    @geometry.setter
    def geometry(self, geometry: Geometry | None):
        if not (geometry is None or isinstance(geometry, Geometry)):
            raise TypeError(
                f"WorldObject.geometry must be a Geometry object or None, not {geometry!r}"
            )
        self._store.geometry = geometry

    @property
    def material(self) -> Material | None:
        """The object's material, the data that defines the appearance of this object."""
        # In contrast to the geometry, the material is not stored on self._store,
        # because it should not be tracked, because pipeline-containers are unique
        # for each combi of (wobject, material, renderstate).
        return self._material

    @material.setter
    def material(self, material: Material | None) -> None:
        if not (material is None or isinstance(material, Material)):
            raise TypeError(
                f"WorldObject.geometry must be a Geometry object or None, not {material!r}"
            )
        self._material = material

    @property
    def cast_shadow(self) -> bool:
        """Whether this object casts shadows, i.e. whether it is rendered into
        a shadow map. Default False."""
        return self._cast_shadow  # does not affect any shaders

    @cast_shadow.setter
    def cast_shadow(self, value: bool) -> None:
        self._cast_shadow = bool(value)

    @property
    def receive_shadow(self) -> bool:
        """Whether this object receives shadows. Default False."""
        return self._store.receive_shadow

    @receive_shadow.setter
    def receive_shadow(self, value: bool) -> None:
        self._store.receive_shadow = bool(value)

    @property
    def parent(self) -> WorldObject | None:
        """Object's parent in the scene graph (read-only).
        An object can have at most one parent.
        """
        if self._parent is None:
            return None
        else:
            return self._parent()

    @property
    def children(self) -> Tuple[WorldObject, ...]:
        """tuple of children of this object. (read-only)"""
        return tuple(self._children)

    def add(
        self,
        *objects: WorldObject,
        before: WorldObject | None = None,
        keep_world_matrix: bool = False,
    ) -> WorldObject:
        """Add child objects.

        Any number of objects may be added. Any current parent on an object
        passed in here will be removed, since an object can have at most one
        parent. If ``before`` argument is given, then the items are inserted
        before the given element.

        Parameters
        ----------
        *objects : WorldObject
            The world objects to add as children.
        before : WorldObject
            If not None, insert the objects before this child object.
        keep_world_matrix : bool
            If True, the child will keep it's world transform. It moves in the
            scene graph but will visually remain in the same place. If False,
            the child will keep it's parent transform.

        """
        for obj in objects:
            if obj.parent is not None:
                obj.parent.remove(obj, keep_world_matrix=keep_world_matrix)

            if before is not None:
                idx = self._children.index(before)
            else:
                idx = len(self._children)

            if keep_world_matrix:
                transform_matrix = obj.world.matrix

            obj._parent = weakref.ref(self)
            obj.world.parent = self.world
            self._children.insert(idx, obj)
            self.world.children.insert(idx, obj.world)

            if keep_world_matrix:
                obj.world.matrix = transform_matrix

        return self

    def remove(self, *objects: WorldObject, keep_world_matrix: bool = False) -> None:
        """Removes object as child of this object. Any number of objects may be removed."""
        for obj in objects:
            try:
                self._children.remove(obj)
                self.world.children.remove(obj.world)
            except ValueError:
                logger.warning("Attempting to remove object that was not a child.")
                continue
            else:
                obj._reset_parent(keep_world_matrix=keep_world_matrix)

    def clear(self, *, keep_world_matrix: bool = False) -> None:
        """Removes all children."""

        for child in self._children:
            child._reset_parent(keep_world_matrix=keep_world_matrix)

        self._children.clear()
        self.world.children.clear()

    def _reset_parent(self, *, keep_world_matrix=False):
        """Sets the parent to None.

        xref: https://github.com/pygfx/pygfx/pull/482#discussion_r1135670771
        """

        if keep_world_matrix:
            transform_matrix = self.world.matrix

        self._parent = None
        self.world.parent = None

        if keep_world_matrix:
            self.world.matrix = transform_matrix

    def traverse(
        self, callback: Callable[[WorldObject], Any], skip_invisible: bool = False
    ):
        """Executes the callback on this object and all descendants.

        If ``skip_invisible`` is given and True, objects whose
        ``visible`` property is False - and their children - are
        skipped. Note that modifying the scene graph inside the callback
        is discouraged.
        """

        for child in self.iter(skip_invisible=skip_invisible):
            callback(child)

    def iter(
        self,
        filter_fn: Callable[[WorldObject], bool] | None = None,
        skip_invisible: bool = False,
    ) -> Iterator[WorldObject]:
        """Create a generator that iterates over this objects and its children.
        If ``filter_fn`` is given, only objects for which it returns ``True``
        are included.
        """
        if skip_invisible and not self.visible:
            return

        if filter_fn is None:
            yield self
        elif filter_fn(self):
            yield self

        for child in self._children:
            yield from child.iter(filter_fn, skip_invisible)

    def _get_bounds_from_geometry(self):
        geometry = self.geometry
        if geometry is None:
            self._bounds_geometry = None
        elif isinstance(positions_buf := getattr(geometry, "positions", None), Buffer):
            if self._bounds_geometry_rev == positions_buf.rev:
                return self._bounds_geometry
            self._bounds_geometry = None
            # Get array and check expected shape
            positions_array = positions_buf.data
            if (
                positions_array is not None
                and positions_array.ndim == 2
                and positions_array.shape[1] in (2, 3)
            ):
                self._bounds_geometry = Bounds.from_points(positions_array)
                self._bounds_geometry_rev = positions_buf.rev
        elif isinstance(grid_buf := getattr(geometry, "grid", None), Texture):
            if self._bounds_geometry_rev == grid_buf.rev:
                return self._bounds_geometry
            # account for multi-channel image data
            grid_shape = tuple(reversed(grid_buf.size[: grid_buf.dim]))
            # create aabb in index/data space
            aabb = np.array([np.zeros_like(grid_shape), grid_shape[::-1]], dtype="f8")
            # convert to local image space by aligning
            # center of voxel index (0, 0, 0) with origin (0, 0, 0)
            aabb -= 0.5
            # ensure coordinates are 3D
            # NOTE: important we do this last, we don't want to apply
            # the -0.5 offset to the z-coordinate of 2D images
            if aabb.shape[1] == 2:
                aabb = np.hstack([aabb, [[0], [0]]])
            self._bounds_geometry = Bounds(aabb, None)
            self._bounds_geometry_rev = grid_buf.rev
        else:
            self._bounds_geometry = None
        return self._bounds_geometry

    def get_geometry_bounding_box(self) -> np.ndarray | None:
        bounds = self._get_bounds_from_geometry()
        if bounds is not None:
            return bounds.aabb

    def get_bounding_box(self) -> np.ndarray | None:
        """Axis-aligned bounding box in local model space.

        Returns
        -------
        aabb : ndarray, [2, 3] or None
            An axis-aligned bounding box, or None when the object does
            not take up a particular space.
        """

        # Collect bounding boxes
        _aabbs = []
        for child in self._children:
            aabb = child.get_bounding_box()
            if aabb is not None:
                trafo = child.local.matrix
                _aabbs.append(la.aabb_transform(aabb, trafo))
        bounds = self._get_bounds_from_geometry()
        if bounds is not None:
            _aabbs.append(bounds.aabb)

        # Combine
        if _aabbs:
            aabbs = np.stack(_aabbs)
            final_aabb = np.zeros((2, 3), dtype=float)
            final_aabb[0] = np.min(aabbs[:, 0, :], axis=0)
            final_aabb[1] = np.max(aabbs[:, 1, :], axis=0)
        else:
            final_aabb = None

        return final_aabb

    def get_bounding_sphere(self) -> np.ndarray | None:
        """Bounding Sphere in local model space.

        Returns
        -------
        bounding_shere : ndarray, [4] or None
            A sphere (x, y, z, radius), or None when the object does
            not take up a particular space.

        """
        # NOTE: this currently does not even use the sphere-data from the geometry!
        aabb = self.get_bounding_box()
        return None if aabb is None else la.aabb_to_sphere(aabb)

    def get_world_bounding_box(self) -> np.ndarray | None:
        """Axis aligned bounding box in world space.

        Returns
        -------
        aabb : ndarray, [2, 3] or None
            The transformed axis-aligned bounding box, or None when the
            object does not take up a particular space.

        """
        aabb = self.get_bounding_box()
        return None if aabb is None else la.aabb_transform(aabb, self.world.matrix)

    def get_world_bounding_sphere(self) -> np.ndarray | None:
        """Bounding Sphere in world space.

        Returns
        -------
        bounding_shere : ndarray, [4] or None
            A sphere (x, y, z, radius), or None when the object does
            not take up a particular space.

        """
        aabb = self.get_world_bounding_box()
        return None if aabb is None else la.aabb_to_sphere(aabb)

    def _wgpu_get_pick_info(self, pick_value) -> dict:
        # In most cases the material handles this.
        return self.material._wgpu_get_pick_info(pick_value)

    def look_at(self, target: WorldObject) -> None:
        """Orient the object so it looks at the given position.

        This sets the object's rotation such that its ``forward`` direction
        points towards ``target`` (given in world space). This rotation takes
        reference_up into account, i.e., the rotation is chosen in such a way that a
        camera looking ``forward`` follows the rotation of a human head looking
        around without tilting the head sideways.

        """

        self.world.forward = target - self.world.position



import numpy as np


class KeyframeTrack:
    def __init__(self, name, target, path, times, values, interpolation):
        self.name = name
        self.target = target
        self.path = path

        assert len(times) == len(values), "times and values must have the same length"

        self._optimize(times, values)
        self.interpolation = interpolation(times, values)

    def _optimize(self, times, values):
        # removes equivalent sequential keys as common in morph target sequences
        # (0,0,0,0,1,1,1,0,0,0,0,0,0,0) --> (0,0,1,1,0,0)
        optimized_times = []
        optimized_values = []
        prev_time = None
        prev_value = None
        for time, value in zip(times, values, strict=True):
            # remove adjacent keyframes scheduled at the same times
            if time != prev_time:
                if np.any(value != prev_value):
                    if prev_time != optimized_times[-1] if optimized_times else None:
                        optimized_times.append(prev_time)
                        optimized_values.append(prev_value)
                    optimized_times.append(time)
                    optimized_values.append(value)
                    prev_value = value

                prev_time = time

        # ensure the last keyframe is added
        if optimized_times[-1] != times[-1]:
            optimized_times.append(times[-1])
            optimized_values.append(values[-1])

        self.times = np.array(optimized_times)
        self.values = np.array(optimized_values)
        return self.times, self.values



"""
All pygfx builtin shaders.

The idea is that these only import from the (public) namespace pygfx.renderers.wgpu.
"""

# ruff: noqa: F401

from . import backgroundshader
from . import gridshader
from . import meshshader
from . import pointsshader
from . import lineshader
from . import imageshader
from . import volumeshader
from . import textshader



from typing import Tuple

import pylinalg as la

from ._orbit import OrbitController


class TrackballController(OrbitController):
    """A controller to freely rotate a camera around a center position.

    This controller is similar to the OrbitController, but it does not
    maintain a constant camera up vector.

    Default controls:

    * Left mouse button: orbit / rotate.
    * Right mouse button: pan.
    * Fourth mouse button: quickzoom
    * wheel: zoom to point.
    * alt+wheel: adjust fov.
    """

    _default_controls = OrbitController._default_controls.copy()

    def rotate(self, delta: Tuple, rect: Tuple, *, animate=False):
        """Rotate around the target, using two angles (in radians).

        If animate is True, the motion is damped. This requires the
        controller to receive events from the renderer/viewport.
        """
        return super().rotate(delta, rect, animate=animate)

    def _update_rotate(self, delta):
        assert isinstance(delta, tuple) and len(delta) == 2

        dx, dy = delta
        camera_state = self._get_camera_state()

        position = camera_state["position"]
        rotation = camera_state["rotation"]

        qx = la.quat_from_axis_angle((0, 1, 0), -dx)
        qy = la.quat_from_axis_angle((1, 0, 0), -dy)

        delta_rot = la.quat_mul((0, 0, 0, 1), la.quat_mul(qy, qx))

        # Get rotations
        rot1 = rotation
        rot2 = la.quat_mul(rot1, delta_rot)

        # Calculate new position
        pos1 = position
        pos2target1 = self._get_target_vec(camera_state, rotation=rot1)
        pos2target2 = self._get_target_vec(camera_state, rotation=rot2)
        pos2 = pos1 + pos2target1 - pos2target2

        # Apply new state
        new_camera_state = {"position": pos2, "rotation": rot2}
        self._set_camera_state(new_camera_state)



"""
Implements the base shader class. The shader is responsible for providing the
WGSL code, as well as providing the information to connect it to the resources
(buffers and textures) and some details on the pipeline and rendering.
"""

from ....resources import Buffer, Texture
from ..engine.utils import (
    GfxSampler,
    GfxTextureView,
    to_vertex_format,
    to_texture_format,
    hash_from_value,
)
from ..engine.binding import Binding
from .resolve import resolve_shadercode
from .bindings import BindingDefinitions
from .templating import apply_templating


class ShaderInterface:
    """Define what a shader object must look like from the pov of the pipeline."""

    def __init__(self):
        self._hash = None

    @property
    def hash(self):
        """A hash of the current state of the shader. If the hash changed, it's likely that the shader changed."""
        if self._hash is None:
            return self._get_hash()
        else:
            return self._hash

    def _get_hash(self):
        raise NotImplementedError()

    def generate_wgsl(self, **template_vars):
        raise NotImplementedError()

    def get_bindings_info(self, wobject, shared):
        """Subclasses must return a dict describing the buffers and
        textures used by this shader.

        The result must be a dict of dicts with binding objects
        (group_slot -> binding_slot -> binding)
        """
        return {
            0: {},
        }

    def get_pipeline_info(self, wobject, shared):
        """Subclasses must return a dict describing pipeline details.

        Fields for a compute shader: empty

        Fields for a render shader:
          * "cull_mode"
          * "primitive_topology"
        """
        return {
            "primitive_topology": 0,
            "cull_mode": 0,
        }

    def get_render_info(self, wobject, shared):
        """Subclasses must return a dict describing render details.

        Fields for a compute shader:
          * "indices" (3 ints)

        Fields for a render shader:
          * "indices" (list of 2 or 4 ints).
        """
        return {
            "indices": (1, 1),
        }


class BaseShader(ShaderInterface):
    """Base shader object to compose and template shaders using jinja2.

    Templating variables can be passed kwargs, set(and get) as attributes, or passed as kwargs to ``generate_wgsl()``.
    They should only be set in __init__() and get_bindings_info().

    The template variables affect the hash. After `pipeline.py` calls get_bindings_info(), it checks
    the hash to see if the shader has changed, and recompiles the wgsl if it has.
    """

    type = "render"  # must be "compute" or "render"
    needs_bake_function = False

    def __init__(self, wobject, **template_vars):
        super().__init__()

        # The shader templating variables
        self._template_vars_init = {}
        self._template_vars_init.update(template_vars)
        self._template_vars_bindings = {}

        # Start off with the current dict set. This is reset when get_bindings_info() gets called.
        self._template_vars_current = self._template_vars_init

        # Handling binding definitions is handled by a wrapped object.
        self._binding_definitions = BindingDefinitions()

        # Init variables that apply to all world objects. The wobject can be None, mostly for testing.
        if wobject is not None:
            self["n_clipping_planes"] = wobject.material.clipping_plane_count
            self["clipping_mode"] = wobject.material.clipping_mode
            self["use_alpha_test"] = wobject.material._gfx_use_alpha_test
            self["alpha_compare"] = wobject.material.alpha_compare

        # Init other common variables so we don't need jinja2's defined()
        self["colormap_dim"] = None

    def __setitem__(self, key, value):
        if self._template_vars_current is None:
            raise RuntimeError(
                "Attempt to set template variable outside of _init() or _get_bindings()."
            )
        self._template_vars_current[key] = value

    def __getitem__(self, key):
        try:
            return self._template_vars_bindings[key]
        except KeyError:
            return self._template_vars_init[key]

    def _get_hash(self):
        # Getting the hash also locks template var setting
        self._template_vars_current = None

        # The full name of this shader class.
        fullname = self.__class__.__module__ + "." + self.__class__.__name__

        # If we assume that the shader class produces the same code for
        # a specific set of template-vars, we can use the fullname in the hash
        # instead of the actual code. This assumption is valid in
        # general, but can break down in a few specific situations, the
        # most common one being an interactive session. In this case,
        # the fullname would be something like "__main__.CustomShader".
        # To be on the safe side, we use the full code when the fullname
        # contains only one dot. This may introduce false positives,
        # but that's fine, because this is only a performance
        # optimization.

        name_probably_defines_code = fullname.count(".") >= 2

        if name_probably_defines_code:
            # Faster, but assumes that the produced code only depends on template-vars.
            return hash_from_value(
                [
                    fullname,
                    self._binding_definitions.get_code(),
                    self._template_vars_init,
                    self._template_vars_bindings,
                ]
            )
        else:
            # More reliable (e.g. in an interactive session).
            return hash_from_value(
                [
                    self._binding_definitions.get_code(),
                    self.get_code(),
                    self._template_vars_init,
                    self._template_vars_bindings,
                ]
            )

    def get_bindings_info(self, wobject, shared):
        # We assume that in normal usage (by the pipeline.py logic),
        # this method is called first after initialization,
        # which means that we can handle the _template_vars_current here.
        self._template_vars_current = self._template_vars_bindings

        # Clear the binding-based template variables, and the binding definitions, so that when get_bindings()
        # not-sets a variable (which was previously set), this affects the hash and is detected.
        self._template_vars_bindings.clear()
        self._binding_definitions.clear()

        try:
            return self.get_bindings(wobject, shared)
        finally:
            self._template_vars_current = None
            self._hash = None  # template vars my have changed, so force a recalculation

    def get_bindings(self, wobject, shared):
        # Default implementation returns zero bindings
        return {0: {}}

    def get_code(self):
        """Implement this to compose the total (but still templated)
        shader. This method is called by ``generate_wgsl()``.
        """
        raise NotImplementedError()

    def generate_wgsl(self, **more_template_vars):
        """Generate the final WGSL. Calls get_code() and then resolves
        the templating variables, varyings, and depth output.
        """

        # Compose shader variables
        template_vars = {}
        template_vars.update(self._template_vars_init)
        template_vars.update(self._template_vars_bindings)
        template_vars.update(more_template_vars)
        template_vars["bindings_code"] = self._binding_definitions.get_code()

        # If template variables are used in ``get_code()``, make sure its correct.
        ori_template_vars, self._template_vars_bindings = (
            self._template_vars_bindings,
            template_vars,
        )

        try:
            # Obtain base code
            code1 = self.get_code()
            # Templating and resolving includes
            code2 = apply_templating(code1, **template_vars)
            # Resolve Varyings and FragmentOutput
            code3 = resolve_shadercode(code2)
            return code3
        finally:
            self._template_vars_bindings = ori_template_vars

    def define_bindings(self, bindgroup, bindings_dict):
        """Define a collection of bindings organized in a dict."""
        if self._template_vars_current is not self._template_vars_bindings:
            raise RuntimeError("define_bindings can only be used from get_bindings")

        for index, binding in bindings_dict.items():
            self._binding_definitions.define_binding(bindgroup, index, binding)

    def define_binding(self, bindgroup, index, binding):
        """Define a uniform, buffer, sampler, or texture.

        The binding must be a Binding object. The code that defines the binding
        will be inserted in ``pygfx.std.wgsl``.
        """
        if self._template_vars_current is not self._template_vars_bindings:
            raise RuntimeError("define_bindings can only be used from get_bindings")

        self._binding_definitions.define_binding(bindgroup, index, binding)

    # ----- Colormap stuff

    def define_generic_colormap(self, map, texcoords):
        """Define the given texture as the colormap to be used to
        lookup the final color from the (per-vertex or per-face) texcoords. In
        the WGSL the colormap can be sampled using ``sample_colormap()``.
        Returns a list of bindings.

        For colormaps in mesh, line, points. Supports 1D/2D/3D
        textures, different texture formats.
        """

        filter_mode = f"{map.mag_filter}, {map.min_filter}, {map.mipmap_filter}"
        address_mode = f"{map.wrap_s}, {map.wrap_t}"

        sampler = GfxSampler(filter_mode, address_mode)
        texture_view = self._define_colormap_texture(map.texture)

        # Check that texture dim matches texcoords
        if not isinstance(texcoords, Buffer):
            raise ValueError("texture is present, but texcoords must be a buffer")
        vert_fmt = to_vertex_format(texcoords.format)
        if texture_view.view_dim == "1d" and "x" not in vert_fmt:
            pass
        elif not vert_fmt.endswith("x" + texture_view.view_dim[0]):
            raise ValueError(
                f"texcoords {texcoords.format} does not match texture_view {texture_view.view_dim}"
            )

        # Return bindings
        return [
            Binding("s_colormap", "sampler/filtering", sampler, "FRAGMENT"),
            Binding("t_colormap", "texture/auto", texture_view, "FRAGMENT"),
            Binding("u_colormap", "buffer/uniform", map.uniform_buffer, "FRAGMENT"),
        ]

    def define_img_colormap(self, map):
        """Define the given texture view as the colormap to be used to
        lookup the final color from the image data.
        In the WGSL the colormap can be sampled using ``sample_colormap()``.
        Returns a list of bindings.
        """

        filter_mode = f"{map.mag_filter}, {map.min_filter}, {map.mipmap_filter}"
        address_mode = f"{map.wrap_s}, {map.wrap_t}"

        sampler = GfxSampler(filter_mode, address_mode)
        texture_view = self._define_colormap_texture(map.texture)

        # Check that texture dim matches image channels
        if int(texture_view.view_dim[0]) != self["img_nchannels"]:
            raise ValueError(
                f"Image channels {self['img_nchannels']} does not match texture_view {texture_view.view_dim}"
            )

        # Return bindings
        return [
            Binding("s_colormap", "sampler/filtering", sampler, "FRAGMENT"),
            Binding("t_colormap", "texture/auto", texture_view, "FRAGMENT"),
            Binding("u_colormap", "buffer/uniform", map.uniform_buffer, "FRAGMENT"),
        ]

    def _define_colormap_texture(self, texture):
        # Get texture view
        if not isinstance(texture, Texture):
            raise TypeError("texture must be a Texture")
        texture_view = GfxTextureView(texture)
        # Dimensionality
        view_dim = texture_view.view_dim
        if view_dim not in ("1d", "2d", "3d"):
            raise ValueError(f"Unexpected texture dimension: '{view_dim}'")
        self["colormap_dim"] = view_dim
        self["colormap_coord_type"] = {
            "1d": "f32",
            "2d": "vec2<f32>",
            "3d": "vec3<f32>",
        }[view_dim]
        # Sampling type
        fmt = to_texture_format(texture_view.format)
        if "norm" in fmt or "float" in fmt:
            self["colormap_format"] = "f32"
        elif "uint" in fmt:
            self["colormap_format"] = "u32"
        else:
            self["colormap_format"] = "i32"
        # Channels
        self["colormap_nchannels"] = len(fmt) - len(fmt.lstrip("rgba"))
        return texture_view



from typing import Tuple

import pylinalg as la

from ._base import Controller


class FlyController(Controller):
    """A controller to fly around a scene as if it's a flight simulator.

    Default controls:

    * Left mouse button: rotate.
    * "wasd": move forward, backwards, and strafe to the sides.
    * space/shift: move up/down.
    * "qe": roll the camera/aircraft around it's axis.
    * wheel: increase/decrease maximum speed.
    * Fourth mouse button: quickzoom
    * alt+wheel: adjust fov.
    """

    _default_controls = {
        "mouse1": ("rotate", "drag", (0.005, 0.005)),
        "q": ("roll", "repeat", -2),
        "e": ("roll", "repeat", +2),
        "w": ("move", "repeat", (0, 0, -1)),
        "s": ("move", "repeat", (0, 0, +1)),
        "a": ("move", "repeat", (-1, 0, 0)),
        "d": ("move", "repeat", (+1, 0, 0)),
        " ": ("move", "repeat", (0, +1, 0)),
        "shift": ("move", "repeat", (0, -1, 0)),
        "mouse4": ("quickzoom", "peek", 2),
        "wheel": ("speed", "push", -0.001),
        "alt+wheel": ("fov", "push", -0.01),
    }

    def __init__(self, camera, *, speed=None, **kwargs):
        super().__init__(camera, **kwargs)

        if speed is None:
            cam_state = camera.get_state()
            approx_scene_size = 0.5 * (cam_state["width"] + cam_state["height"])
            scene_fly_thru_time = 5  # seconds
            speed = approx_scene_size / scene_fly_thru_time
        self.speed = speed

    @property
    def speed(self):
        """The (maximum) speed that the camera will move, in units per second.
        By default it's based off the width and height of the camera.
        """
        return self._speed

    @speed.setter
    def speed(self, value):
        self._speed = float(value)

    def rotate(self, delta: Tuple, rect: Tuple, *, animate=False):
        """Rotate the camera over two angles (pitch and yaw in radians).

        If animate is True, the motion is damped. This requires the
        controller to receive events from the renderer/viewport.
        """
        if animate:
            action_tuple = ("rotate", "push", (1.0, 1.0))
            action = self._create_action(None, action_tuple, (0.0, 0.0), None, rect)
            action.set_target(delta)
            action.snap_distance = 0.01
            action.done = True
        elif self._cameras:
            self._update_rotate(delta)
            return self._update_cameras()

    def _update_rotate(self, delta):
        assert isinstance(delta, tuple) and len(delta) == 2

        dx, dy = delta
        camera_state = self._get_camera_state()
        rotation = camera_state["rotation"]

        qx = la.quat_from_axis_angle((0, 1, 0), -dx)
        qy = la.quat_from_axis_angle((1, 0, 0), -dy)

        delta_rot = la.quat_mul((0, 0, 0, 1), la.quat_mul(qy, qx))

        new_rotation = la.quat_mul(rotation, delta_rot)

        # Apply new state
        new_camera_state = {"rotation": new_rotation}
        self._set_camera_state(new_camera_state)

    def roll(self, delta: float, rect: Tuple, *, animate=False):
        """Rotate the camera over the z-axis (roll, in radians).

        If animate is True, the motion is damped. This requires the
        controller to receive events from the renderer/viewport.
        """
        if animate:
            action_tuple = ("roll", "push", 1.0)
            action = self._create_action(None, action_tuple, 0.0, None, rect)
            action.set_target(delta)
            action.snap_distance = 0.01
            action.done = True
        elif self._cameras:
            self._update_roll(delta)
            return self._update_cameras()

    def _update_roll(self, delta):
        assert isinstance(delta, float)

        camera_state = self._get_camera_state()
        rotation = camera_state["rotation"]

        qz = la.quat_from_axis_angle((0, 0, 1), -delta)
        new_rotation = la.quat_mul(rotation, qz)

        new_camera_state = {"rotation": new_rotation}
        self._set_camera_state(new_camera_state)

    def move(self, delta: Tuple, rect: Tuple, *, animate=False):
        """Move the camera in the given (x, y, z) direction.

        The delta is expressed in the camera's local coordinate frame.
        Forward is in -z direction, because as (per the gltf spec) a
        camera looks down it's negative Z-axis.

        If animate is True, the motion is damped. This requires the
        controller to receive events from the renderer/viewport.
        """

        if animate:
            action_tuple = ("move", "push", (1.0, 1.0, 1.0))
            action = self._create_action(
                None, action_tuple, (0.0, 0.0, 0.0), None, rect
            )
            action.set_target(delta)
            action.done = True
        elif self._cameras:
            self._update_move(delta)
            return self._update_cameras()

    def _update_move(self, delta):
        assert isinstance(delta, tuple) and len(delta) == 3

        cam_state = self._get_camera_state()
        position = cam_state["position"]
        rotation = cam_state["rotation"]
        delta_world = la.vec_transform_quat(delta, rotation)

        new_position = position + delta_world * self.speed
        self._set_camera_state({"position": new_position})

    def _update_speed(self, delta):
        assert isinstance(delta, float)
        speed = self.speed * 2**delta
        self.speed = max(0.001, speed)



from ..renderers import Renderer


class Viewport:
    """A rectangular area on a renderer.

    Parameters
    ----------
    renderer : Renderer
        The renderer on which the viewport should be defined.
    rect : tuple, [4]
        The viewport rect (x, y, w, h). If None, it is set to the full size of
        the renderer's canvas.

    """

    def __init__(self, renderer, rect=None):
        if not isinstance(renderer, Renderer):
            raise TypeError("Viewport first arg must be a Renderer.")
        self._renderer = renderer
        self.rect = rect

    @classmethod
    def from_viewport_or_renderer(cls, x):
        """Convenience method. Some parts of pygfx that accept a viewport
        might just as well accept a renderer.
        """
        if isinstance(x, Viewport):
            return x
        elif isinstance(x, Renderer):
            return cls(x)
        else:
            raise TypeError("Expected a Viewport or a Renderer.")

    @property
    def renderer(self):
        """A reference to the renderer that this is a viewport of."""
        return self._renderer

    @property
    def logical_size(self):
        """The logical size of the rectangular area described by this viewport."""
        rect = self.rect
        return rect[2], rect[3]

    @property
    def rect(self):
        """The viewport rect (x, y, w, h). Can be set to None to set
        it to the full renderer size.
        """
        return self._get_rect()

    @rect.setter
    def rect(self, rect):
        if rect is None:
            self._rect = None
        else:
            if not (len(rect) == 4):
                raise ValueError("Viewport rect must be 4 numbers, or None.")
            self._rect = tuple(float(v) for v in rect)

    def _get_rect(self):
        """Can be overloaded, e.g. for viewports that define their own rect."""
        if self._rect is None:
            return (0, 0, *self.renderer.logical_size)
        else:
            return self._rect

    def is_inside(self, x, y):
        """Get whether the given position is inside the viewport rect."""
        vp = self.rect
        return vp[0] <= x <= vp[0] + vp[2] and vp[1] <= y <= vp[1] + vp[3]

    def render(self, scene, camera, flush=False):
        """A shorthand for ``renderer.render(scene, camera)`` at the appropriate
        viewport. Does not flush by default.
        """
        self.renderer.render(scene, camera, rect=self.rect, flush=flush)



from importlib.util import find_spec

import numpy as np

from ._base import Geometry


def geometry_from_trimesh(mesh):
    """Convert a Trimesh geometry object to pygfx geometry.

    Creates a Geometry object from the given `trimesh.Trimesh
    <https://trimsh.org/trimesh.html#trimesh.Trimesh>`_ object.

    Parameters
    ----------
    mesh : Trimesh
        The mesh to be converted into a geometry.

    Returns
    -------
    converted_mesh : Geometry
        A Geometry object representing the given mesh.

    """
    from trimesh import Trimesh

    if not isinstance(mesh, Trimesh):
        raise NotImplementedError(f"No support for trimesh.{mesh.__class__.__name__}")

    kwargs = dict(
        positions=np.ascontiguousarray(mesh.vertices, dtype="f4"),
        indices=np.ascontiguousarray(mesh.faces, dtype="i4"),
        normals=np.ascontiguousarray(mesh.vertex_normals, dtype="f4"),
    )
    # Note: some trimesh visuals have type texture but no UV coordinates
    if mesh.visual.kind == "texture" and getattr(mesh.visual, "uv", None) is not None:
        # convert the uv coordinates from opengl to wgpu conventions.
        # wgpu uses the D3D and Metal coordinate systems.
        # the coordinate origin is in the upper left corner, while the opengl coordinate
        # origin is in the lower left corner.
        # trimesh loads textures according to the opengl coordinate system.
        wgpu_uv = mesh.visual.uv * np.array([1, -1]) + np.array(
            [0, 1]
        )  # uv.y = 1 - uv.y
        kwargs["texcoords"] = np.ascontiguousarray(wgpu_uv, dtype="f4")
    elif mesh.visual.kind == "vertex":
        kwargs["colors"] = np.ascontiguousarray(mesh.visual.vertex_colors, dtype="f4")
    elif mesh.visual.kind == "face":
        kwargs["colors"] = np.ascontiguousarray(mesh.visual.face_colors, dtype="f4")

    # todo: support vertex attribute 'tangent'

    return Geometry(**kwargs)


def geometry_from_open3d(x) -> Geometry:
    """
    Convert an Open3D geometry object to a pygfx Geometry.

    This function handles the conversion of Open3D geometry objects such as
    `open3d.geometry.TriangleMesh <https://www.open3d.org/docs/release/python_api/open3d.geometry.TriangleMesh.html>`_
    or `open3d.geometry.PointCloud <https://www.open3d.org/docs/release/python_api/open3d.geometry.PointCloud.html>`_
    into pygfx Geometry objects. Depending on the input type, it delegates the conversion
    to specialized conversion functions.

    Parameters
    ----------
    x : open3d.geometry.TriangleMesh or open3d.geometry.PointCloud
        The geometry to be converted into a pygfx Geometry.

    Returns
    -------
    converted_geometry : Geometry
        A Geometry object representing the given Open3D geometry.

    Raises
    ------
    ImportError
        If the open3d library is not installed.
    NotImplementedError
        If the input geometry type is not supported.
    """
    if not find_spec("open3d"):
        raise ImportError(
            "The `open3d` library is required for this function: pip install open3d"
        )

    from open3d import geometry as o3d_geometry

    if isinstance(x, o3d_geometry.TriangleMesh):
        return geometry_from_open3d_triangle_mesh(x)
    elif isinstance(x, o3d_geometry.PointCloud):
        return geometry_from_open3d_point_cloud(x)
    else:
        raise NotImplementedError(
            "Conversion for the provided Open3D geometry type is not implemented."
        )


def geometry_from_open3d_triangle_mesh(x) -> Geometry:
    """
    Convert an Open3D TriangleMesh object to a pygfx Geometry.

    This function creates a pygfx Geometry object from the given
    `open3d.geometry.TriangleMesh <https://www.open3d.org/docs/release/python_api/open3d.geometry.TriangleMesh.html>`_
    by extracting vertices, triangles, normals, and texture coordinates (if available).

    Parameters
    ----------
    x : open3d.geometry.TriangleMesh
        The TriangleMesh object to be converted into a pygfx Geometry.

    Returns
    -------
    converted_mesh : Geometry
        A Geometry object representing the given TriangleMesh.

    Raises
    ------
    ImportError
        If the open3d library is not installed.
    NotImplementedError
        If the input is not an instance of open3d.geometry.TriangleMesh.
    """
    if not find_spec("open3d"):
        raise ImportError(
            "The `open3d` library is required for this function: pip install open3d"
        )

    from open3d import geometry as o3d_geometry

    if not isinstance(x, o3d_geometry.TriangleMesh):
        raise NotImplementedError(
            "Input must be an instance of open3d.geometry.TriangleMesh."
        )

    vertices = np.ascontiguousarray(x.vertices, dtype=np.float32)
    triangles = np.ascontiguousarray(x.triangles, dtype="i4")

    kwargs = dict(positions=vertices, indices=triangles)

    # Add normals if available
    if len(x.vertex_normals) > 0:
        kwargs["normals"] = np.ascontiguousarray(x.vertex_normals, dtype=np.float32)

    # Add UV coordinates if available
    if len(x.triangle_uvs) > 0:
        triangle_uvs = np.ascontiguousarray(x.triangle_uvs, dtype=np.float32)

        vertex_uvs = np.zeros((len(vertices), 2), np.float32)
        vertex_uvs[triangles.flat] = triangle_uvs

        # Adjust UVs for rendering systems
        vertex_uvs_wgpu = (vertex_uvs * np.array([1, -1]) + np.array([0, 1])).astype(
            np.float32
        )
        kwargs["texcoords"] = vertex_uvs_wgpu

    return Geometry(**kwargs)


def geometry_from_open3d_point_cloud(x) -> Geometry:
    """
    Convert an Open3D PointCloud object to a pygfx Geometry.

    This function creates a pygfx Geometry object from the given
    `open3d.geometry.PointCloud <https://www.open3d.org/docs/release/python_api/open3d.geometry.PointCloud.html>`_
    by extracting points, colors, and normals (if available).

    Parameters
    ----------
    x : open3d.geometry.PointCloud
        The PointCloud object to be converted into a pygfx Geometry.

    Returns
    -------
    converted_geometry : Geometry
        A Geometry object representing the given PointCloud.

    Raises
    ------
    ImportError
        If the open3d library is not installed.
    NotImplementedError
        If the input is not an instance of open3d.geometry.PointCloud.
    """
    if not find_spec("open3d"):
        raise ImportError(
            "The `open3d` library is required for this function: pip install open3d"
        )

    from open3d import geometry as o3d_geometry

    if not isinstance(x, o3d_geometry.PointCloud):
        raise NotImplementedError(
            "Input must be an instance of open3d.geometry.PointCloud."
        )

    points = np.ascontiguousarray(x.points, dtype=np.float32)
    kwargs = dict(positions=points)

    # Add colors if available
    if len(x.colors) > 0:
        kwargs["colors"] = np.ascontiguousarray(x.colors, dtype=np.float32)

    # Add normals if available
    if len(x.normals) > 0:
        kwargs["normals"] = np.ascontiguousarray(x.normals, dtype=np.float32)

    return Geometry(**kwargs)



import numpy as np

from ._base import id_provider
from . import WorldObject, Mesh, Line
from ..resources import Buffer

DOCSTRING_TEMPLATE = """Display a {name} multiple times using instances.

    An instanced {name} with a matrix for each instance.

    Parameters
    ----------
    geometry : Geometry
        The {name}'s geometry data.
    material : Material
        The material with which to render the {name}.
    count : int
        The number of instances to create.
    kwargs : Any
        Additional kwargs get forwarded to the :class:`base class
        <pygfx.objects.{base_cls}>`.

    """


class InstancedObject(WorldObject):
    __doc__ = DOCSTRING_TEMPLATE.format(name="object", base_cls="WorldObject").replace(
        "a object", "an object"
    )

    def __init__(self, geometry, material, count, **kwargs):
        super().__init__(geometry, material, **kwargs)
        count = int(count)
        # Create array of `count` instance_info objects
        dtype = np.dtype(
            [
                ("matrix", np.float32, (4, 4)),
                ("global_id", np.uint32),
                ("_12_bytes_padding", np.uint8, (12,)),
            ]
        )
        instance_infos = np.zeros(count, dtype)
        self._store.instance_buffer = Buffer(
            instance_infos, nitems=count, force_contiguous=True
        )
        # Set ids
        self._idmap = {}
        for instance_index in range(count):
            id = id_provider.claim_id(self)
            self._idmap[id] = instance_index
            instance_infos[instance_index]["global_id"] = id
        # Init eye matrices
        for i in range(4):
            instance_infos["matrix"][:, i, i] = 1

    def __del__(self):
        super().__del__()
        instance_infos = self._store["instance_buffer"].data
        for i in range(len(instance_infos)):
            id_provider.release_id(self, instance_infos[i]["global_id"])

    @property
    def instance_buffer(self):
        return self._store.instance_buffer

    def set_matrix_at(self, index: int, matrix):
        """set the matrix for the instance at the given index."""
        matrix = np.array(matrix).reshape(4, 4)
        self._store["instance_buffer"].data["matrix"][index] = matrix.T
        self._store["instance_buffer"].update_range(index, 1)

    def get_matrix_at(self, index: int):
        """get the matrix for the instance at the given index."""
        return self._store["instance_buffer"].data["matrix"][index].T

    def _wgpu_get_pick_info(self, pick_value) -> dict:
        info = super()._wgpu_get_pick_info(pick_value)
        # The id maps to one of our instances
        id = pick_value & 1048575  # 2**20-1
        info["instance_index"] = self._idmap.get(id)
        return info


class InstancedMesh(Mesh, InstancedObject):
    __doc__ = DOCSTRING_TEMPLATE.format(name="mesh", base_cls="Mesh")

    def __init__(self, geometry, material, count, **kwargs):
        super().__init__(geometry, material, count, **kwargs)


class InstancedLine(Line, InstancedObject):
    __doc__ = DOCSTRING_TEMPLATE.format(name="line", base_cls="Line")

    def __init__(self, geometry, material, count, **kwargs):
        super().__init__(geometry, material, count, **kwargs)



import numpy as np
import pylinalg as la

from ._base import Geometry
from ._plane import generate_plane
from .utils import merge


def box_geometry(
    width=1,
    height=1,
    depth=1,
    width_segments=1,
    height_segments=1,
    depth_segments=1,
):
    """Generate a box (rectangular cuboid).

    Creates a box (a rectangular cuboid) of the given size that is centered
    around the local frame's origin. Faces may be subdivided by specifying the
    number of segments along each axis. This will result in all faces parallel to
    the given axis to be evenly divided into the requested number of segments.

    Parameters
    ----------
    width : int
        Size along the x-axis.
    height : int
        Size along the y-axis.
    depth : int
        Size along the z-axis.
    width_segments : int
        Number of segments along x-axis.
    height_segments : int
        Number of segments along y-axis.
    depth_segments : int
        Number of segments along z-axis.

    Returns
    -------
    box : Geometry
        A geometry object containing the requested box shape.
        Mathematically, it consists of a set of open orientable manifolds.

    """

    cube_dim = np.array([width, height, depth], dtype=np.float32)
    cube_seg = np.array(
        [width_segments, height_segments, depth_segments], dtype=np.uint32
    )
    cube_normal_up = np.array(
        [
            [[1, 0, 0], [0, 1, 0]],  # right
            [[-1, 0, 0], [0, 1, 0]],  # left
            [[0, 1, 0], [1, 0, 0]],  # top
            [[0, -1, 0], [-1, 0, 0]],  # bottom
            [[0, 0, 1], [0, 1, 0]],  # front (this matches the default plane)
            [[0, 0, -1], [0, 1, 0]],  # back
        ],
        dtype=np.float32,
    )

    plane_csys = np.array(
        [
            *cube_normal_up[4],
            np.cross(
                np.ndarray.astype(cube_normal_up[4, 0], np.float32, copy=False),
                np.ndarray.astype(cube_normal_up[4, 1], np.float32, copy=False),
            ),
        ]
    )

    planes = []
    for normal, up in cube_normal_up:
        normal = np.ndarray.astype(normal, np.float32, copy=False)
        up = np.ndarray.astype(up, np.float32, copy=False)

        plane_idx = np.flatnonzero(normal == 0)
        (
            plane_positions,
            plane_normals,
            plane_texcoords,
            plane_index,
        ) = generate_plane(*cube_dim[plane_idx], *cube_seg[plane_idx])

        affine = np.identity(4, dtype=np.float32)

        sign_idx = np.flatnonzero(normal != 0)[0]
        affine[:-1, -1] = (cube_dim[sign_idx] / 2) * normal

        swap_axes = np.dot(
            np.array([normal, up, np.cross(normal, up)]).T,
            plane_csys,
        )

        if normal[2] == 0:
            theta = np.pi / 2
            rotate90 = np.array(
                [
                    [np.cos(theta), -np.sin(theta), 0],
                    [np.sin(theta), np.cos(theta), 0],
                    [0, 0, 1],
                ]
            )
            affine[:-1, :-1] = np.dot(
                swap_axes,
                rotate90,
            )
        else:
            affine[:-1, :-1] = swap_axes

        plane_positions = la.vec_transform(plane_positions, affine)

        affine[:-1, -1] = 0
        plane_normals = la.vec_transform(plane_normals, affine)

        planes.append((plane_positions, plane_normals, plane_texcoords, plane_index))

    positions, normals, texcoords, indices = merge(planes)
    positions = positions.astype(np.float32)
    normals = normals.astype(np.float32)

    return Geometry(
        indices=indices, positions=positions, normals=normals, texcoords=texcoords
    )



from ._base import Material
from ..utils import Color
from ..utils.enums import CoordSpace


class GridMaterial(Material):
    """A Cartesian grid.

    Implements "the best darn grid shader yet" a.k.a. the Prestine grid by Ben Golus.

    Parameters
    ----------
    major_step : 2-tuple | float
        The step distance between the major ticks, for the first and second dimension, respectively.
        Default (1.0, 1.0).
    minor_step : 2-tuple | float
        The step distance between the minor ticks, for the first and second dimension, respectively.
        Default (0.1, 0.1).
    axis_thickness : float
        The thickness of the axis lines. Default 0.0 (no axis lines).
    major_thickness : float
        The thickness of the major grid lines. Default 2.0.
    minor_thickness : float
        The thickness of the minor grid lines. Default 0.0 (no minor grid lines).
    thickness_space : str | CoordSpace
        The coordinate space in which the thickness is expressed ('screen' or 'world'). Default 'screen'.
    axis_color : str | tuple | Color
       The color of the axis lines. Default '#777'.
    major_color : str | tuple | Color
       The color of the major grid lines. Default '#777'.
    minor_color : str | tuple | Color
       The color of the minor grid lines. Default '#777'.
    infinite : bool
        Whether the grid is infinite. Default True.
    alpha_test : float
        The alpha test of a grid is 0.1 by default, so that only the lines are visible, and the 'holes' don't hide other objects.
    kwargs : Any
        Additional kwargs are passed to the base constructor
        (:class:`pygfx.materials.Material`).
    """

    uniform_type = dict(
        Material.uniform_type,
        major_step="2xf4",
        minor_step="2xf4",
        axis_thickness="f4",
        major_thickness="f4",
        minor_thickness="f4",
        axis_color="4xf4",
        major_color="4xf4",
        minor_color="4xf4",
    )

    def __init__(
        self,
        *,
        major_step=(1.0, 1.0),
        minor_step=(0.1, 0.1),
        axis_thickness=0.0,
        major_thickness=2.0,
        minor_thickness=0.0,
        thickness_space="screen",
        axis_color="#777",
        major_color="#777",
        minor_color="#777",
        infinite=True,
        alpha_test=0.1,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.major_step = major_step
        self.minor_step = minor_step
        self.axis_thickness = axis_thickness
        self.major_thickness = major_thickness
        self.minor_thickness = minor_thickness
        self.thickness_space = thickness_space
        self.axis_color = axis_color
        self.major_color = major_color
        self.minor_color = minor_color
        self.infinite = infinite
        self.alpha_test = alpha_test

    def _set_draw_axis(self):
        self._store.draw_axis = self.axis_thickness > 0

    def _set_draw_major(self):
        thickness = self.major_thickness
        step = self.major_step
        self._store.draw_major = thickness > 0 and (step[0] > 0 or step[1] > 0)

    def _set_draw_minor(self):
        thickness = self.minor_thickness
        step = self.minor_step
        self._store.draw_minor = thickness > 0 and (step[0] > 0 or step[1] > 0)

    @property
    def _gfx_draw_axis(self):
        return self._store.draw_axis

    @property
    def _gfx_draw_major(self):
        return self._store.draw_major

    @property
    def _gfx_draw_minor(self):
        return self._store.draw_minor

    @property
    def major_step(self):
        """The step distance between the major grid lines."""
        return tuple(float(x) for x in self.uniform_buffer.data["major_step"])

    @major_step.setter
    def major_step(self, step):
        if isinstance(step, (float, int)):
            step = max(0.0, float(step))
            step = step, step
        if isinstance(step, (tuple, list)) and len(step) == 2:
            step = max(0.0, float(step[0])), max(0.0, float(step[1]))
        else:
            raise TypeError(
                f"major_step must be tuple or float, not {step.__class__.__name__}"
            )
        self.uniform_buffer.data["major_step"] = step
        self.uniform_buffer.update_full()
        self._set_draw_major()

    @property
    def minor_step(self):
        """The step distance between the minor grid lines."""
        return tuple(float(x) for x in self.uniform_buffer.data["minor_step"])

    @minor_step.setter
    def minor_step(self, step):
        if isinstance(step, (float, int)):
            step = max(0.0, float(step))
            step = step, step
        if isinstance(step, (tuple, list)) and len(step) == 2:
            step = max(0.0, float(step[0])), max(0.0, float(step[1]))
        else:
            raise TypeError(
                f"minor_step must be tuple or float, not {step.__class__.__name__}"
            )
        self.uniform_buffer.data["minor_step"] = step
        self.uniform_buffer.update_full()
        self._set_draw_minor()

    @property
    def axis_thickness(self):
        """The thickness of the axis lines."""
        return float(self.uniform_buffer.data["axis_thickness"])

    @axis_thickness.setter
    def axis_thickness(self, thickness):
        self.uniform_buffer.data["axis_thickness"] = max(0.0, float(thickness))
        self.uniform_buffer.update_full()
        self._set_draw_axis()

    @property
    def major_thickness(self):
        """The thickness of the major grid lines."""
        return float(self.uniform_buffer.data["major_thickness"])

    @major_thickness.setter
    def major_thickness(self, thickness):
        self.uniform_buffer.data["major_thickness"] = max(0.0, float(thickness))
        self.uniform_buffer.update_full()
        self._set_draw_major()

    @property
    def minor_thickness(self):
        """The thickness of the minor grid lines."""
        return float(self.uniform_buffer.data["minor_thickness"])

    @minor_thickness.setter
    def minor_thickness(self, thickness):
        self.uniform_buffer.data["minor_thickness"] = max(0.0, float(thickness))
        self.uniform_buffer.update_full()
        self._set_draw_minor()

    @property
    def thickness_space(self):
        """The coordinate space in which the thicknesses are expressed.

        See :obj:`pygfx.utils.enums.CoordSpace`:
        """
        return self._store.thickness_space

    @thickness_space.setter
    def thickness_space(self, value):
        value = value or "screen"
        if value not in CoordSpace:
            raise ValueError(
                f"GridMaterial.thickness_space must be a string in {CoordSpace}, not {value!r}"
            )
        elif value not in (CoordSpace.world, CoordSpace.screen):
            raise ValueError(
                f"GridMaterial.thickness_space must be a either 'screen'  or 'world', not '{value}'."
            )
        self._store.thickness_space = value

    @property
    def axis_color(self):
        """The color of the axis lines."""
        return Color(self.uniform_buffer.data["axis_color"])

    @axis_color.setter
    def axis_color(self, color):
        self.uniform_buffer.data["axis_color"] = Color(color)
        self.uniform_buffer.update_full()

    @property
    def major_color(self):
        """The color of the major grid lines."""
        return Color(self.uniform_buffer.data["major_color"])

    @major_color.setter
    def major_color(self, color):
        self.uniform_buffer.data["major_color"] = Color(color)
        self.uniform_buffer.update_full()

    @property
    def minor_color(self):
        """The color of the minor grid lines."""
        return Color(self.uniform_buffer.data["minor_color"])

    @minor_color.setter
    def minor_color(self, color):
        self.uniform_buffer.data["minor_color"] = Color(color)
        self.uniform_buffer.update_full()

    @property
    def infinite(self):
        """Whether the grid is infinite.

        If not infinite, the grid is 1x1 in world space, scaled, rotated, and
        positioned with the object's transform.

        (Infinite grids are not actually infinite. Rather they move along with
        the camera, and are sized based on the distance between the camera and
        the grid.)
        """
        return self._store.infinite

    @infinite.setter
    def infinite(self, value):
        self._store.infinite = bool(value)



import numpy as np

from ._base import Geometry


def klein_bottle_geometry(scale=1.0, stitch=False):
    """Generate a Klein bottle.

    The Klein bottle is a surface for which the inside and outside
    are the same, similar to a Möbius strip. In fact, a Klein bottle
    can be constructed by glueing together two Möbius strips.

    Parameters
    ----------
    scale : float
        The scale of the bottle.
    stitch : bool
        Whether to stitch the ends together to produce a closed
        manifold. If True, a mathematically correct Klein bottle is
        produced. If False (default) an approximation is produced where
        the ends of the bottle meet, but are not actually connected.

    Returns
    -------
    klein_bottle : Geometry
        A geometry object representing the requested klein bottle.
        Mathematically, it is either an "orientable open manifold" or
        a "non-orientable closed manifold", depending on the ``stitch``
        parameter.

    """

    # This is an interesting object for mathematicians. For us it's
    # interesting because we can test whether our lighting etc. deals
    # correctly with objects for which the "inside" must also be shown.

    # The number of vertices is nxn
    n = 40

    # Get 2D surface in 3D space
    u = np.linspace(0, 2 * np.pi, n, endpoint=True, dtype=np.float32)
    v = np.linspace(0, 2 * np.pi, n, endpoint=False, dtype=np.float32)
    ux, vx = np.meshgrid(u, v)
    x, y, z = klein_bottle_surface(ux, vx)

    # Scaled to a unit cube, then scale to width / height / depth
    # x = (x + 1.66559) * (0.0437597 * width)
    # y = (y - 2.04939) * (0.0277017 * height)
    # z = (z + 0.00000) * (0.0833333 * depth)
    # Scaled to fit inside a unit cube, but maintain original proportions
    x = (x + 1.66559) * (0.0833333 * scale)
    y = (y - 2.04939) * (0.0833333 * scale)
    z = (z + 0.00000) * (0.0833333 * scale)

    # Put into an Nx4 array
    positions = np.empty((x.size, 3), np.float32)
    positions[:, 0] = x.flat
    positions[:, 1] = y.flat
    positions[:, 2] = z.flat

    # Texcoords are easy
    texcoords = np.column_stack([ux.flat, vx.flat]).astype(np.float32, copy=False)
    texcoords *= 1 / (2 * np.pi)

    # Map indices
    # Two triangles onto the "top-left" rectangle (six vertices)
    indices = np.array([0, 1, n + 1, n + 1, n, 0], np.uint32)
    # Replicate to all rectangles, add offsets
    indices = np.tile(indices, (n, n - 1, 1))
    gx, gy = np.meshgrid(
        np.arange(indices.shape[1], dtype=np.uint32),
        n * np.arange(indices.shape[0], dtype=np.uint32),
    )
    indices += (gx + gy).reshape((*indices.shape[:2], 1))

    # Stitch the faces at the tube's edge.
    indices[-1, :, 2:5] -= n * n

    if stitch:
        # We can do the same for the tube's ends, and make this a closed
        # manifold! Note that in doing that, we also make it
        # not-orientable - the orientation of the faces (i.e. the
        # winding) switches where we apply the stitch. Also note that
        # in the current implementation we're left with n unused
        # vertices.

        # In the code below, i1 are one end of the tube, and i2 the
        # other. Since the tube is inside-out, the matching pairs are
        # on opposing sides.
        for i in range(n):
            i2 = n - 1 + i * n
            i1 = (n // 2 - i) * n
            i1 = i1 if i1 >= 0 else i1 + positions.shape[0]
            indices[indices == i2] = i1

    # Create buffers for this geometry
    indices = indices.reshape((-1, 3))
    return Geometry(indices=indices, positions=positions, texcoords=texcoords)


def klein_bottle_surface(u, v):
    """
    http://paulbourke.net/geometry/toroidal/

        A German topologist named Klein
        Thought the Möbius Loop was divine
        Said he, "If you glue
        The edges of two
        You get a weird bottle like mine.
    """
    half = (0 <= u) & (u < np.pi)
    r = 4 * (1 - np.cos(u) / 2)
    x = 6 * np.cos(u) * (1 + np.sin(u)) + r * np.cos(v + np.pi)
    x[half] = (6 * np.cos(u) * (1 + np.sin(u)) + r * np.cos(u) * np.cos(v))[half]
    y = 16 * np.sin(u)
    y[half] = (16 * np.sin(u) + r * np.sin(u) * np.cos(v))[half]
    z = r * np.sin(v)
    return x, y, z


def torus_knot_geometry(
    scale=1.0, tube=0.4, tubular_segments=64, radial_segments=8, p=2, q=3, stitch=False
):
    """Generate a torus knot.

    Create geometry representing a torus knot, the particular shape of which is
    defined by a pair of coprime integers, p and q. If p and q are not coprime,
    the result will be a torus link.

    Parameters
    ----------
    scale : float
        The scale of the torus, default 1.
    tube : float
        The radius of the tube. Default 0.4.
    tubular_segments : int
        default is 64.
    radial_segments : int
        default is 8.
    p : int
        How many times the geometry winds around its axis of
        rotational symmetry. Default 2.
    q : int
        How many times the geometry winds around a circle in
        the interior of the torus. Default 3.
    stitch : bool
        Whether to stitch the ends together to produce a closed
        manifold. Default False. If False, the mesh is basically a
        curved surface with the edges meeting to make it visually
        closed, which works better for texturing. Set to True for a
        mathematically closed object.

    Returns
    -------
    torus : Geometry
        A geometry object representing the requested torus.
        Mathematically, it is an open orientable manifold, which can
        be closed with the ``stitch`` parameter.

    """

    # If we do not stitch, the two ends meet (i.e. duplicate vertices)
    # to make the mesh look closed (while mathematically it is not).

    if stitch:
        tubular_verts = tubular_segments
        radial_verts = radial_segments
    else:
        tubular_verts = tubular_segments + 1
        radial_verts = radial_segments + 1

    # Define base factors
    u = np.linspace(
        0, p * 2 * np.pi, tubular_verts, endpoint=not stitch, dtype=np.float32
    )
    v = np.linspace(0, 2 * np.pi, radial_verts, endpoint=not stitch, dtype=np.float32)

    # Get positions along the torus' center, and a tiny step further
    pos1 = torus_knot_surface(u, p, q, scale)
    pos2 = torus_knot_surface(u + 0.01, p, q, scale)

    # Two vectors along the torus' centerline
    vec1 = np.ndarray.astype(pos1 - pos2, np.float32, copy=False)
    vec2 = np.ndarray.astype(pos1 + pos2, np.float32, copy=False)

    # Two vectors orthoginal to the torus' centerline
    vec3 = np.cross(vec1, vec2)
    vec4 = np.cross(vec3, vec1)
    # Normalize
    vec3 /= ((vec3[:, 0] ** 2 + vec3[:, 1] ** 2 + vec3[:, 2] ** 2) ** 0.5).reshape(
        -1, 1
    )
    vec4 /= ((vec4[:, 0] ** 2 + vec4[:, 1] ** 2 + vec4[:, 2] ** 2) ** 0.5).reshape(
        -1, 1
    )
    # Define positions relative to the centerline
    cx = -tube * np.cos(v)
    cy = +tube * np.sin(v)
    # Prepare shapes, so we can do numpy broadcast
    pos = pos1.reshape(-1, 1, 3)
    cx.shape = 1, -1, 1
    cy.shape = 1, -1, 1
    vec3.shape = -1, 1, 3
    vec4.shape = -1, 1, 3
    # Broadcast!
    positions = pos + cx * vec4 + cy * vec3
    normals = positions - pos
    positions.shape = -1, 3
    normals.shape = -1, 3
    normals *= 1 / np.linalg.norm(normals, axis=1).reshape(-1, 1)

    # Create texcords
    # ty, tx = np.meshgrid(u / u[-1], v / v[-1])
    ty, tx = np.meshgrid(v / v[-1], u / u[-1])
    texcoords = np.column_stack((tx.flat, ty.flat))

    # Create indices
    # Two triangles onto the "top-left" rectangle (six vertices)
    if stitch:
        base_triangle = [radial_verts - 1, 0, radial_verts, radial_verts, 0, 1]
    else:
        base_triangle = [radial_verts, 0, radial_verts + 1, radial_verts + 1, 0, 1]
    base_triangle = np.array(base_triangle, np.uint32)

    # Replicate to all rectangles, add offsets
    indices = np.tile(base_triangle, (tubular_segments, radial_segments, 1))
    gx, gy = np.meshgrid(
        np.arange(indices.shape[1], dtype=np.uint32),
        radial_verts * np.arange(indices.shape[0], dtype=np.uint32),
    )
    indices += (gx + gy).reshape((*indices.shape[:2], 1))

    # Correct the faces at the tube's ends
    if stitch:
        indices[indices >= len(positions)] -= len(positions)

    indices = indices.reshape((-1, 3))
    # indices = np.fliplr(indices)  # Use this to change winding between CW and CCW

    return Geometry(
        indices=indices, positions=positions, normals=normals, texcoords=texcoords
    )


def torus_knot_surface(u, p, q, radius):
    """Taken from ThreeJS, but vectorized."""
    cu = np.cos(u)
    su = np.sin(u)
    qu_over_p = q / p * u
    cs = np.cos(qu_over_p)
    x = (2 + cs) * cu * (0.5 * radius)
    y = (2 + cs) * su * (0.5 * radius)
    z = np.sin(qu_over_p) * (0.5 * radius)
    return np.column_stack((x, y, z))



"""Objects to control cameras.

.. currentmodule:: pygfx.controllers

Controllers define how cameras can be interacted with. They are
independent of any GUI toolkits.

.. autosummary::
    :toctree: controllers/
    :template: ../_templates/custom_layout.rst

    Controller
    PanZoomController
    OrbitController
    TrackballController
    FlyController

"""

from ._base import Controller
from ._panzoom import PanZoomController
from ._orbit import OrbitController
from ._trackball import TrackballController
from ._fly import FlyController


__all__ = [
    "Controller",
    "FlyController",
    "OrbitController",
    "PanZoomController",
    "TrackballController",
]



"""The Pygfx render engine."""

# ruff: noqa: F401, F403

from ._version import __version__, version_info, repo_dir as _repo_dir
from . import utils

from .resources import *
from .objects import *
from .geometries import *
from .materials import *
from .cameras import *
from .helpers import *
from .controllers import *
from .animation import *

from .renderers import *

from .utils.color import Color
from .utils.load import load_mesh, load_meshes, load_scene
from .utils.load_gltf import (
    load_gltf,
    load_gltf_async,
    load_gltf_mesh,
    load_gltf_mesh_async,
    print_scene_graph,
)
from .utils.show import show, Display
from .utils.viewport import Viewport
from .utils.text import font_manager
from .utils import cm, enums, logger
from .utils.enums import *

# Temp fix for pyinstaller to pick up pylinalg
import pylinalg

del pylinalg


def _get_dependency_version_ranges():
    # The only case where this dependency checking makes sense is for devs
    # using pygfx from a Git repo.
    if not _repo_dir:
        return {}
    # Try import, dont care when this fails (e.g. frozen or py < 3.11)
    try:
        import os, tomllib  # noqa
    except ImportError:
        return {}
    # Load dependency versions
    with open(os.path.join(_repo_dir, "pyproject.toml"), "rb") as fp:
        dependencies = tomllib.load(fp)["project"]["dependencies"]
    # Parse
    limits_per_dependency = {}
    for name_verlimits in dependencies:
        name, _, verlimits = name_verlimits.partition(" ")
        if name in ("wgpu", "pylinalg"):
            min_ver = max_ver = 0
            for lim in verlimits.split(","):
                if lim.startswith(">="):
                    min_ver = tuple(map(int, lim[2:].split(".")))
                elif lim.startswith("<"):
                    max_ver = tuple(map(int, lim[1:].split(".")))
            if min_ver and max_ver:
                limits_per_dependency[name] = min_ver, max_ver
    return limits_per_dependency


def _check_lib_version(libname, pipname):
    import importlib

    if libname not in _dependency_version_ranges:
        return
    lib = importlib.import_module(libname)
    min_ver, max_ver = _dependency_version_ranges[libname]
    detected = f"Detected {lib.__version__}, need >={min_ver}, <{max_ver}."
    if lib.version_info < min_ver:
        logger.error(
            f"Incompatible version of {libname}:\n    {detected}\n    To update, use e.g. `pip install -U {pipname}`."
        )
    elif lib.version_info >= max_ver:
        logger.warning(f"Possible incompatible version of {libname}:\n    {detected}")


_dependency_version_ranges = _get_dependency_version_ranges()
_check_lib_version("wgpu", "wgpu")
_check_lib_version("pylinalg", "pylinalg")


def _get_sg_image_scraper():
    """Hook for sphinx so we can tell it how to generate the gallery."""
    import sphinx_gallery.scrapers
    from .utils.gallery_scraper import pygfx_scraper

    # add webp as supported extension
    sphinx_gallery.scrapers._KNOWN_IMG_EXTS += ("webp",)

    return pygfx_scraper


# Elements of this library are derived from three.js, original license
# at time of writing copied here:
# ---
# The MIT License
#
# Copyright © 2010-2022 three.js authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ---
# End of original license copy.



"""
Implements the base classes for trackable classes.

## Introduction

This module is implemented in a more or less generic way. The code does not
refer to world objects, geometry, materials or resources. But don't be
fooled - this code is *very* much tailored to the needs of such objects.
In particular:

* we want to track properties (that may need to trigger a pipeline).
* we want to be able to track objects themselves depending on the situation
  (because the pipeline only cares about the object's properties while the
  bindings much match to the correct buffers/textures).
* we want to be able to swap out similar objects and not trigger of the
  new object has the same properties as before (so that we can e.g. replace
  colormaps or texture atlasses).
* we want to be able to track multiple trees independently.

The last two points mean that the code must be aware of the tree
structures that trackables make up.

## High level overview

We define a Trackable object as something of which the properties can
be tracked (its properties can be other Trackable objects).

The PropTracker is the object that keeps track of the changes. One
first tracks usage of properties, after which any changes to these
properties get communicated to the tracker.

## Finer details

The Trackable uses a Store - a custom dictionary object that has hooks
for attribute access. This is where the magic happens. The Trackable
object itself is just a container for the store. This has the benefit
that subclasses of Trackable just store the props that they want to
track on its store.

The Store keeps a set of what trackers need to be notified (using weak
refs). The tracker keeps a set of what stores notify it, for the sole
reason that the tracker can disconnect these stores when usage is
re-tracked.

Every store has a unique id, and the tracker uses (id, prop) tuples as
keys to keep track of things. When a property sets/unsets a Trackable,
the tree is resolved at that moment (i.e. the tracker is temporary aware
of the tree). This avoids the need for the tracker (or stores?) to keep
track of a hierarchy which would make the code a lot more complex (I
tried and it was not fun). The only downside (that I can think of) is
that setting `matererial.map = other_map` works while
`matererial.map = None; material.map = other_map` does not, because
there is no persistent knowledge of hierarchy.

"""

import threading
import weakref

global_id_counter = 0
global_lock = threading.RLock()
global_context = None

simple_value_types = None.__class__, bool, int, float, str


def get_comp_value(value):
    if isinstance(value, simple_value_types):
        return value
    elif isinstance(value, tuple):
        return tuple(get_comp_value(v) for v in value)
    else:
        try:
            return "hash:" + str(hash(value))
        except TypeError:
            return "id:" + str(id(value))


class Undefined:
    def __repr__(self):
        return "undefined"


undefined = Undefined()


class Trackable:
    """A base class to make an object trackable."""

    def __init__(self):
        self._store = Store()


class TrackContext:
    """A context used when tracking usage of trackable objects."""

    def __init__(self, tracker, label):
        assert isinstance(tracker, PropTracker)
        assert isinstance(label, str)
        self.tracker = tracker
        self.label = label

    def __enter__(self):
        global global_context
        global_lock.acquire()
        global_context = self
        self.tracker._track_init(self.label)
        return None

    def __exit__(self, value, type, tb):
        global global_context
        self.tracker._track_done(self.label)
        self.tracker = None
        self.label = 0
        global_context = None
        global_lock.release()


class Store(dict):
    """Object to store key-value pairs that are tracked. Each Trackable
    has one such object. Must use attribute access to trigger the tracking
    mechanics. The internals can use index access.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        # Give it an id
        global global_id_counter
        with global_lock:
            global_id_counter += 1
            self["_trackable_id"] = global_id_counter  # f"t{global_id_counter}"
        # Store what trackers we need to notify of changes
        self["_trackable_trackers"] = weakref.WeakSet()

    def __hash__(self):
        return self["_trackable_id"]

    def __setattr__(self, key, new_value):
        # Apply
        old_value = dict.get(self, key, undefined)
        self[key] = new_value
        # Notify the trackers
        id = self["_trackable_id"]
        for tracker in self["_trackable_trackers"]:
            tracker._track_set((id, key), old_value, new_value)

    def __getattribute__(self, key):
        if key.startswith("__"):
            return dict.__getattribute__(self, key)
        value = undefined
        try:
            value = self[key]
            return value
        except KeyError:
            return dict.__getattribute__(self, key)
        finally:
            if global_context:
                global_context.tracker._track_get(
                    self, key, value, global_context.label
                )


class PropTracker:
    """Object to store all the tracking data."""

    def __init__(self):
        # Keep track of what stores have a reference to *this*,
        # so that we can clear that reference when we want.
        # The store and tracker are always connected/disconnected together.
        self._stores = weakref.WeakKeyDictionary()  # store -> labels
        # The names to track. Names are (id, key)  -  name -> labels
        self._trackable_names = {}
        # Keep track of values  -  name -> (ref_value, cur_value)
        self._trackable_values = {}
        # Keep track of what has changed  -  name -> labels
        self._trackable_changed = {}

    def track_usage(self, label):
        """Used to track the usage of attributes. The result of this method
        should be used as a context manager. This method must only be
        used by the one system that is tracking this object's changes.
        """
        return TrackContext(self, label)

    def pop_changed(self):
        """Get a set of labels for which the object (and its child
        trackable objects) has changed. Also resets the changed status
        to unchanged. This method must only be used by the one system
        that is tracking this object's changes.
        """
        # Quick version
        if not self._trackable_changed:
            return set()

        # Reset the cached values
        for name in self._trackable_changed:
            try:
                _ref_value, cur_value = self._trackable_values[name]
            except KeyError:
                pass
            else:
                self._trackable_values[name] = cur_value, cur_value
        # Collect changed label
        result = set()
        for labels in self._trackable_changed.values():
            result.update(labels)
        # Reset and return
        self._trackable_changed = {}
        return set(label.lstrip("!") for label in result)

    def _track_store(self, store, label):
        # Introduce the store and tracker to each-other
        store["_trackable_trackers"].add(self)
        labels = self._stores.setdefault(store, set())
        labels.add(label)

    def _untrack_store(self, store, label):
        # Break the connection between tracker and store
        labels = self._stores.setdefault(store, set())
        labels.discard(label)
        if not labels:
            self._stores.pop(store, None)
            store["_trackable_trackers"].discard(self)

    def _track_init(self, label):
        # Reset the tracking data for the given label
        for name in list(self._trackable_names):
            labels = self._trackable_names[name]
            labels.discard(label)
            if not labels:
                self._trackable_names.pop(name)
                self._trackable_values.pop(name)

        # Break connections with old stores
        to_remove = []
        for store, labels in self._stores.items():
            labels.discard(label)
            if not labels:
                to_remove.append(store)
        for store in to_remove:
            self._stores.pop(store, None)
            store["_trackable_trackers"].discard(self)

    def _track_done(self, label):
        pass

    def _track_get(self, store, key, value, label):
        # Called when *any* trackable has an attribute GET while a
        # global context is active. The trackable can be in the tree
        # of the tracker (a (grand)child), but can also be part of a
        # detached tree.

        self._track_store(store, label)

        id = store["_trackable_id"]
        name = id, key

        self._trackable_names.setdefault(name, set()).add(label)
        comp_value = get_comp_value(value)
        self._trackable_values[name] = comp_value, comp_value

    def _track_set(self, name, old_value, new_value):
        # Called when a trackable, that is setup to notify this tracker,
        # has an attribute SET. When this attribute was itself a
        # trackable, and/or is a trackable, we must update the complete
        # tree behind that trackable (to the extend that we track it).

        is_trackable = 0
        old_store = new_store = None
        if isinstance(old_value, (Store, Trackable)):
            is_trackable |= 1
            old_store = old_value if isinstance(old_value, Store) else old_value._store
        if isinstance(new_value, (Store, Trackable)):
            is_trackable |= 2
            new_store = new_value if isinstance(new_value, Store) else new_value._store

        # Get labels
        labels = self._trackable_names.get(name, None)

        if is_trackable and labels:
            # Register / unregister stores
            if is_trackable & 1:
                for label in labels:
                    self._untrack_store(old_store, label)
            if is_trackable & 2:
                for label in labels:
                    self._track_store(new_store, label)
            # If this was and is a trackable, we only process labels starting with "!"
            if is_trackable == 3 and type(old_value) is type(new_value):
                labels = set(label for label in labels if label.startswith("!"))

            # Follow the hierarchy
            if old_store is not None:
                self._track_set_follow_tree(old_store, new_store)

        # Update changed values
        if labels:
            # Mark changed to begin with
            dirty_labels_for_this_name = self._trackable_changed.setdefault(name, set())
            dirty_labels_for_this_name.update(labels)
            # If the value has not changed from the previous (since the
            # last reset) we can un-mark this name as changed.
            comp_value = get_comp_value(new_value)
            ref_value, cur_value = self._trackable_values[name]
            if comp_value == ref_value:
                dirty_labels_for_this_name.difference_update(labels)
                if not dirty_labels_for_this_name:
                    self._trackable_changed.pop(name)
            elif comp_value != cur_value:
                self._trackable_values[name] = ref_value, comp_value

    def _track_set_follow_tree(self, old_store, new_store):
        # old_store must be a store
        # new_store must be a store or None

        # Get names of stuff we track on the old trackable store
        id = old_store["_trackable_id"]
        sub_names = [n for n in self._trackable_names.keys() if n[0] == id]

        # We need to follow these
        for name1 in sub_names:
            _, key = name1
            try:
                value1 = old_store[key]
            except KeyError:
                continue

            # Try get matching path in the new hierarchy
            value2 = undefined
            if new_store is not None:
                try:
                    name2 = new_store["_trackable_id"], key
                    value2 = new_store[key]
                except (AttributeError, KeyError):
                    pass

            if value2 is not undefined:
                # Rename
                self._trackable_names[name2] = self._trackable_names.pop(name1)
                self._trackable_values[name2] = self._trackable_values.pop(name1)
                try:
                    self._trackable_changed[name2] = self._trackable_changed.pop(name1)
                except KeyError:
                    pass
                # Recurse
                self._track_set(name2, value1, value2)

            else:
                # Unregister
                labels = self._trackable_names.pop(name1, ())
                self._trackable_values.pop(name1, None)
                dirty_labels_for_this_name = self._trackable_changed.setdefault(
                    name1, set()
                )
                dirty_labels_for_this_name.update(labels)
                # Unregister stores
                if isinstance(value1, Trackable):
                    for label in labels:
                        self._untrack_store(value1._store, label)
                # Recurse
                if isinstance(value1, Store):
                    self._track_set_follow_tree(value1, None)
                elif isinstance(value1, Trackable):
                    self._track_set_follow_tree(value1._store, None)
